import os
from pathlib import Path
from typing import Any, Dict, List, Union

import pendulum
from prefect import Flow, Task, apply_map
from prefect.backend import set_key_value
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities import logging

from viadot.task_utils import (
    add_ingestion_metadata_task,
    cleanup_validation_clutter,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_csv,
    df_to_parquet,
    dtypes_to_json_task,
    union_dfs_task,
    update_dtypes_dict,
    write_to_json,
)
from viadot.tasks import (
    AzureDataLakeUpload,
    DownloadGitHubFile,
    GetFlowNewDateRange,
    RunGreatExpectationsValidation,
    SupermetricsToDF,
)

logger = logging.get_logger(__name__)

validation_task = RunGreatExpectationsValidation()


class SupermetricsToADLS(Flow):
    def __init__(
        self,
        name: str,
        ds_id: str,
        ds_accounts: List[str],
        fields: List[str],
        ds_user: str = None,
        ds_segments: List[str] = None,
        date_range_type: str = None,
        start_date: str = None,
        end_date: str = None,
        settings: Dict[str, Any] = None,
        filter: str = None,
        max_rows: int = 1000000,
        max_columns: int = None,
        order_columns: str = None,
        expectation_suite: dict = None,
        evaluation_parameters: dict = None,
        keep_validation_output: bool = False,
        output_file_extension: str = ".parquet",
        local_file_path: str = None,
        adls_file_name: str = None,
        adls_dir_path: str = None,
        overwrite_adls: bool = True,
        if_empty: str = "warn",
        if_exists: str = "replace",
        adls_sp_credentials_secret: str = None,
        max_download_retries: int = 5,
        supermetrics_task_timeout: int = 60 * 30,
        parallel: bool = True,
        tags: List[str] = ["extract"],
        vault_name: str = None,
        check_missing_data: bool = True,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading data from different marketing APIs to a local CSV
        using Supermetrics API, then uploading it to Azure Data Lake.

        Args:
            name (str): The name of the flow.
            ds_id (str): A query parameter passed to the SupermetricsToCSV task
            ds_accounts (List[str]): A query parameter passed to the SupermetricsToCSV task
            ds_user (str): A query parameter passed to the SupermetricsToCSV task
            fields (List[str]): A query parameter passed to the SupermetricsToCSV task
            ds_segments (List[str], optional): A query parameter passed to the SupermetricsToCSV task. Defaults to None.
            date_range_type (str, optional): A query parameter passed to the SupermetricsToCSV task. Defaults to None.
            start_date (str, optional): A query parameter to pass start date to the date range filter. Defaults to None.
            end_date (str, optional): A query parameter to pass end date to the date range filter. Defaults to None.
            settings (Dict[str, Any], optional): A query parameter passed to the SupermetricsToCSV task. Defaults to None.
            filter (str, optional): A query parameter passed to the SupermetricsToCSV task. Defaults to None.
            max_rows (int, optional): A query parameter passed to the SupermetricsToCSV task. Defaults to 1000000.
            max_columns (int, optional): A query parameter passed to the SupermetricsToCSV task. Defaults to None.
            order_columns (str, optional): A query parameter passed to the SupermetricsToCSV task. Defaults to None.
            expectation_suite (dict, optional): The Great Expectations suite used to valiaate the data. Defaults to None.
            evaluation_parameters (str, optional): A dictionary containing evaluation parameters for the validation. Defaults to None.
            keep_validation_output (bool, optional): Whether to keep the output files generated by the Great Expectations task. Defaults to False.
            Currently, only GitHub URLs are supported. Defaults to None.
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_file_name (str, optional): Name of file in ADLS. Defaults to None.
            output_file_extension (str, optional): Output file extension - to allow selection of .csv for data which is not easy to handle with parquet. Defaults to ".parquet"..
            adls_dir_path (str, optional): Azure Data Lake destination folder/catalog path. Defaults to None.
            sep (str, optional): The separator to use in the CSV. Defaults to "\t".
            overwrite_adls (bool, optional): Whether to overwrite the file in ADLS. Defaults to True.
            if_empty (str, optional): What to do if the Supermetrics query returns no data. Defaults to "warn".
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
            max_download_retries (int, optional): How many times to retry the download. Defaults to 5.
            supermetrics_task_timeout (int, optional): The timeout for the download task. Defaults to 60*30.
            parallel (bool, optional): Whether to parallelize the downloads. Defaults to True.
            tags (List[str], optional): Flow tags to use, eg. to control flow concurrency. Defaults to ["extract"].
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            check_missing_data (bool, optional): Whether to check missing data. Defaults to True.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        if not ds_user:
            try:
                ds_user = PrefectSecret("SUPERMETRICS_DEFAULT_USER").run()
            except ValueError as e:
                msg = "Neither 'ds_user' parameter nor 'SUPERMETRICS_DEFAULT_USER' secret were not specified"
                raise ValueError(msg) from e

        self.flow_name = name
        self.check_missing_data = check_missing_data
        self.timeout = timeout
        # SupermetricsToDF
        self.ds_id = ds_id
        self.ds_accounts = ds_accounts
        self.ds_segments = ds_segments
        self.ds_user = ds_user
        self.fields = fields
        self.date_range_type = date_range_type
        self.start_date = start_date
        self.end_date = end_date
        self.settings = settings
        self.filter = filter
        self.max_rows = max_rows
        self.max_columns = max_columns
        self.order_columns = order_columns
        self.if_exists = if_exists
        self.output_file_extension = output_file_extension

        # RunGreatExpectationsValidation
        self.expectation_suite = expectation_suite
        self.expectations_path = "/home/viadot/tmp/expectations"
        self.expectation_suite_name = expectation_suite["expectation_suite_name"]
        self.evaluation_parameters = evaluation_parameters
        self.keep_validation_output = keep_validation_output

        # AzureDataLakeUpload
        self.local_file_path = (
            local_file_path or self.slugify(name) + self.output_file_extension
        )
        self.local_json_path = self.slugify(name) + ".json"
        self.now = str(pendulum.now("utc"))
        self.adls_dir_path = adls_dir_path

        if adls_file_name is not None:
            self.adls_file_path = os.path.join(adls_dir_path, adls_file_name)
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", Path(adls_file_name).stem + ".json"
            )

        else:
            self.adls_file_path = os.path.join(
                adls_dir_path, self.now + self.output_file_extension
            )
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", self.now + ".json"
            )
        self.overwrite_adls = overwrite_adls
        self.if_empty = if_empty
        self.adls_sp_credentials_secret = adls_sp_credentials_secret

        # Global
        self.max_download_retries = max_download_retries
        self.supermetrics_task_timeout = supermetrics_task_timeout
        self.parallel = parallel
        self.tags = tags
        self.vault_name = vault_name

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_supermetrics_task(
        self, ds_accounts: Union[str, List[str]], flow: Flow = None
    ) -> Task:
        supermetrics_to_df_task = SupermetricsToDF(timeout=self.timeout)
        t = supermetrics_to_df_task.bind(
            ds_id=self.ds_id,
            ds_accounts=ds_accounts,
            ds_segments=self.ds_segments,
            ds_user=self.ds_user,
            fields=self.fields,
            date_range_type=self.date_range_type,
            start_date=self.start_date,
            end_date=self.end_date,
            settings=self.settings,
            filter=self.filter,
            max_rows=self.max_rows,
            max_columns=self.max_columns,
            order_columns=self.order_columns,
            if_empty=self.if_empty,
            max_retries=self.max_download_retries,
            timeout=self.supermetrics_task_timeout,
            flow=flow,
        )
        return t

    def gen_flow(self) -> Flow:
        if self.check_missing_data is True:
            if self.date_range_type is not None and "days" in self.date_range_type:
                prefect_get_new_date_range = GetFlowNewDateRange(timeout=self.timeout)
                self.date_range_type = prefect_get_new_date_range.run(
                    flow_name=self.flow_name,
                    date_range_type=self.date_range_type,
                    flow=self,
                )

        if self.parallel:
            # generate a separate task for each account
            dfs = apply_map(self.gen_supermetrics_task, self.ds_accounts, flow=self)
            df = union_dfs_task.bind(dfs, flow=self)
        else:
            df = self.gen_supermetrics_task(ds_accounts=self.ds_accounts, flow=self)

        write_json = write_to_json.bind(
            dict_=self.expectation_suite,
            path=os.path.join(
                self.expectations_path, self.expectation_suite_name + ".json"
            ),
            flow=self,
        )

        validation = validation_task.bind(
            df=df,
            expectations_path=self.expectations_path,
            expectation_suite_name=self.expectation_suite_name,
            evaluation_parameters=self.evaluation_parameters,
            keep_output=self.keep_validation_output,
            flow=self,
        )

        if not self.keep_validation_output:
            validation_cleanup = cleanup_validation_clutter.bind(
                expectations_path=self.expectations_path, flow=self
            )
            validation_cleanup.set_upstream(validation, flow=self)
            validation_upstream = validation_cleanup
        else:
            validation_upstream = validation

        df_with_metadata = add_ingestion_metadata_task.bind(df, flow=self)
        dtypes_dict = df_get_data_types_task.bind(df_with_metadata, flow=self)

        df_to_be_loaded = df_map_mixed_dtypes_for_parquet(
            df_with_metadata, dtypes_dict, flow=self
        )

        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df_to_be_loaded,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )

        dtypes_updated = update_dtypes_dict(dtypes_dict, flow=self)
        dtypes_to_json_task.bind(
            dtypes_dict=dtypes_updated, local_json_path=self.local_json_path, flow=self
        )
        json_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        json_to_adls_task.bind(
            from_path=self.local_json_path,
            to_path=self.adls_schema_file_dir_file,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )

        write_json.set_upstream(df, flow=self)
        validation.set_upstream(write_json, flow=self)
        df_with_metadata.set_upstream(validation_upstream, flow=self)
        df_to_be_loaded.set_upstream(dtypes_dict, flow=self)
        dtypes_dict.set_upstream(df_with_metadata, flow=self)

        dtypes_to_json_task.set_upstream(dtypes_updated, flow=self)
        file_to_adls_task.set_upstream(df_to_file, flow=self)
        json_to_adls_task.set_upstream(dtypes_to_json_task, flow=self)
        set_key_value(key=self.adls_dir_path, value=self.adls_file_path)



import pytest
from pyrfc import Connection

from viadot.sources import SAPBW
from viadot.task_utils import credentials_loader

CREDENTIALS = credentials_loader.run(credentials_secret="SAP")
SAPBW = SAPBW(credentials=CREDENTIALS.get("BW"))


@pytest.fixture(scope="session")
def mdx_query_variable():
    mdx_query = """
        SELECT
                {
            }
                ON COLUMNS,
        NON EMPTY
                { 
                    { [0CALMONTH].[202301] } 
        } 
        DIMENSION PROPERTIES
        DESCRIPTION,
        MEMBER_NAME
        ON ROWS

        FROM ZCSALORD1/ZBW4_ZCSALORD1_006_BOA
    
                """
    yield mdx_query


@pytest.fixture(scope="session")
def wrong_mdx_query_variable():
    wrong_mdx_query = """
        SELECT
                {
            }
                ON COLUMNS,
        NON EMPTY
                { 
                    { [0CALMONTH].[202301]  *
        } 
        DIMENSION PROPERTIES
        DESCRIPTION,
        MEMBER_NAME
        ON ROWS

        FROM ZCSALORD1/ZBW4_ZCSALORD1_006_BOA
    
                """
    yield wrong_mdx_query


def test_get_connection():
    conn = SAPBW.get_connection()
    assert isinstance(conn, Connection)


def test_get_all_available_columns(mdx_query_variable):
    all_available_columns = SAPBW.get_all_available_columns(mdx_query_variable)
    assert isinstance(all_available_columns, list)


def test_get_output_data(mdx_query_variable):
    query_output = SAPBW.get_output_data(mdx_query_variable)
    assert isinstance(query_output, dict)


def test_wrong_mdx_query_for_get_output_data(wrong_mdx_query_variable):
    query_output = SAPBW.get_output_data(wrong_mdx_query_variable)
    assert (
        query_output.get("RETURN").get("MESSAGE")
        == "Syntax error: Syntax Error : ... { [0CALMONTH].[2023, row 226, item: 226"
    )



import logging
import os

import pandas as pd
import pyarrow as pa
import pytest

from viadot.signals import SKIP
from viadot.sources.base import SQL, Source

from .test_credentials import get_credentials

logger = logging.getLogger(__name__)


CREDENTIALS = get_credentials("SQL_SOURCE_TEST")
TABLE = "test"
PATH = "t.csv"


class NotEmptySource(Source):
    def to_df(self, if_empty):
        df = pd.DataFrame.from_dict(
            data={"country": ["italy", "germany", "spain"], "sales": [100, 50, 80]}
        )
        return df


class EmptySource(Source):
    def to_df(self, if_empty):
        df = pd.DataFrame()
        if df.empty:
            self._handle_if_empty(if_empty)
        return df


def test_empty_source_skip():
    empty = EmptySource()
    result = empty.to_csv(path=PATH, if_empty="skip")
    assert result is False


def test_to_csv():
    src = NotEmptySource()
    res = src.to_csv(path="testbase.csv")
    assert res == True
    assert os.path.isfile("testbase.csv") == True
    os.remove("testbase.csv")


def test_to_arrow():
    src = NotEmptySource()
    res = src.to_arrow("testbase.arrow")
    assert isinstance(res, pa.Table) == True


def test_to_excel():
    src = NotEmptySource()
    res = src.to_excel(path="testbase.xlsx")
    assert res == True
    assert os.path.isfile("testbase.xlsx") == True
    os.remove("testbase.xlsx")


def test_to_parquet(caplog):
    src = NotEmptySource()
    if os.path.isfile("testbase.parquet") == True:
        os.remove("testbase.parquet")
    with caplog.at_level(logging.INFO):
        src.to_parquet(path="testbase.parquet")
        assert "File not found." in caplog.text
    assert os.path.isfile("testbase.parquet") == True
    os.remove("testbase.parquet")


def test_handle_if_empty(caplog):
    src = EmptySource()
    src._handle_if_empty(if_empty="warn")
    assert "WARNING The query produced no data." in caplog.text
    with pytest.raises(ValueError):
        src._handle_if_empty(if_empty="fail")
    with pytest.raises(SKIP):
        src._handle_if_empty(if_empty="skip")


# def test_to_csv_append():
#     """Test whether `to_csv()` with the append option writes data of correct shape"""
#     driver = "/usr/lib/x86_64-linux-gnu/odbc/libsqlite3odbc.so"
#     db_name = "testfile.sqlite"
#     server = "localhost"
#     source = SQL(
#         credentials=dict(driver=driver, db_name=db_name, server=server, user=None)
#     )

#     # Generate test table.
#     df = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
#     source.create_table("test", dtypes={"a": "INT", "b": "INT"}, if_exists="replace")
#     source.insert_into(TABLE, df)

#     # Write the table to a CSV three times in `append` mode.
#     for i in range(3):
#         source.to_csv(path=PATH, query="SELECT * FROM test", if_exists="append")

#     # Read the CSV and validate no. of rows and columns.
#     out_df = pd.read_csv(PATH, sep="\t")

#     target_length = 3 * df.shape[0]
#     target_width = df.shape[0]

#     actual_length = out_df.shape[0]
#     actual_width = out_df.shape[1]

#     assert actual_length == target_length and actual_width == target_width

#     # Clean up.
#     os.remove(PATH)


# GitHub changes the string and makes the test fail
# def test_conn_str():
#     s = SQL(
#         driver=CREDENTIALS["driver"],
#         server=CREDENTIALS["server"],
#         db=CREDENTIALS["db_name"],
#         user=CREDENTIALS["user"],
#         pw=CREDENTIALS["password"],
#     )
#     assert (
#         s.conn_str
#         == "DRIVER=ODBC Driver 17 for SQL Server;SERVER=s123.database.windows.net;DATABASE=a-b-c;UID={my_user@example.com};PWD={a123;@4}"
#     )



from collections import OrderedDict

from viadot.sources import SAPRFC, SAPRFCV2

sap = SAPRFC()
sap2 = SAPRFCV2()

sql1 = "SELECT a AS a_renamed, b FROM table1 WHERE table1.c = 1"
sql2 = "SELECT a FROM fake_schema.fake_table WHERE a=1 AND b=2 OR c LIKE 'a%' AND d IN (1, 2) LIMIT 5 OFFSET 3"
sql3 = "SELECT b FROM c WHERE testORword=1 AND testANDword=2 AND testLIMITword=3 AND testOFFSETword=4"
sql4 = "SELECT c FROM d WHERE testLIMIT = 1 AND testOFFSET = 2 AND LIMITtest=3 AND OFFSETtest=4"
sql5 = sql3 + " AND longword123=5"
sql6 = "SELECT a FROM fake_schema.fake_table WHERE a=1 AND b=2 OR c LIKE 'a%' AND d IN (1, 2) AND longcolname=3 AND otherlongcolname=5 LIMIT 5 OFFSET 3"
sql7 = """
SELECT a, b
FROM b
WHERE c = 1
AND d = 2
AND longcolname = 12345
AND otherlongcolname = 6789 
AND thirdlongcolname = 01234
LIMIT 5
OFFSET 10
"""


def test__get_table_name():
    assert sap._get_table_name(sql1) == "table1"
    assert sap._get_table_name(sql2) == "fake_schema.fake_table", sap._get_table_name(
        sql2
    )
    assert sap._get_table_name(sql7) == "b"


def test__get_columns():
    assert sap._get_columns(sql1) == ["a", "b"]
    assert sap._get_columns(sql1, aliased=True) == ["a_renamed", "b"], sap._get_columns(
        sql1, aliased=True
    )
    assert sap._get_columns(sql2) == ["a"]
    assert sap._get_columns(sql7) == ["a", "b"]


def test__get_where_condition():
    assert sap._get_where_condition(sql1) == "table1.c = 1", sap._get_where_condition(
        sql1
    )
    assert (
        sap._get_where_condition(sql2) == "a=1 AND b=2 OR c LIKE 'a%' AND d IN (1, 2)"
    ), sap._get_where_condition(sql2)
    assert (
        sap._get_where_condition(sql3)
        == "testORword=1 AND testANDword=2 AND testLIMITword=3 AND testOFFSETword=4"
    ), sap._get_where_condition(sql3)
    assert (
        sap._get_where_condition(sql4)
        == "testLIMIT = 1 AND testOFFSET = 2 AND LIMITtest=3 AND OFFSETtest=4"
    ), sap._get_where_condition(sql4)
    assert (
        sap._get_where_condition(sql7)
        == "c = 1 AND d = 2 AND longcolname = 12345 AND otherlongcolname = 6789"
    ), sap._get_where_condition(sql7)


def test__get_limit():
    assert sap._get_limit(sql1) is None
    assert sap._get_limit(sql2) == 5
    assert sap._get_limit(sql7) == 5


def test__get_offset():
    assert sap._get_offset(sql1) is None
    assert sap._get_offset(sql2) == 3
    assert sap._get_offset(sql7) == 10


def test_client_side_filters_simple():
    _ = sap._get_where_condition(sql5)
    assert sap.client_side_filters == OrderedDict(
        {"AND": "longword123=5"}
    ), sap.client_side_filters


def test_client_side_filters_with_limit_offset():
    _ = sap._get_where_condition(sql6)
    assert sap.client_side_filters == OrderedDict(
        {"AND": "otherlongcolname=5"}
    ), sap.client_side_filters

    _ = sap._get_where_condition(sql7)
    assert sap.client_side_filters == OrderedDict(
        {"AND": "thirdlongcolname = 01234"}
    ), sap.client_side_filters


def test___build_pandas_filter_query():
    _ = sap._get_where_condition(sql6)
    assert (
        sap._build_pandas_filter_query(sap.client_side_filters)
        == "otherlongcolname == 5"
    ), sap._build_pandas_filter_query(sap.client_side_filters)
    _ = sap._get_where_condition(sql7)
    assert (
        sap._build_pandas_filter_query(sap.client_side_filters)
        == "thirdlongcolname == 01234"
    ), sap._build_pandas_filter_query(sap.client_side_filters)


def test__get_table_name_v2():
    assert sap2._get_table_name(sql1) == "table1"
    assert sap2._get_table_name(sql2) == "fake_schema.fake_table", sap2._get_table_name(
        sql2
    )
    assert sap2._get_table_name(sql7) == "b"


def test__get_columns_v2():
    assert sap2._get_columns(sql1) == ["a", "b"]
    assert sap2._get_columns(sql1, aliased=True) == [
        "a_renamed",
        "b",
    ], sap2._get_columns(sql1, aliased=True)
    assert sap2._get_columns(sql2) == ["a"]
    assert sap2._get_columns(sql7) == ["a", "b"]


def test__get_where_condition_v2():
    assert sap2._get_where_condition(sql1) == "table1.c = 1", sap2._get_where_condition(
        sql1
    )
    assert (
        sap2._get_where_condition(sql2) == "a=1 AND b=2 OR c LIKE 'a%' AND d IN (1, 2)"
    ), sap2._get_where_condition(sql2)
    assert (
        sap2._get_where_condition(sql3)
        == "testORword=1 AND testANDword=2 AND testLIMITword=3 AND testOFFSETword=4"
    ), sap2._get_where_condition(sql3)
    assert (
        sap2._get_where_condition(sql4)
        == "testLIMIT = 1 AND testOFFSET = 2 AND LIMITtest=3 AND OFFSETtest=4"
    ), sap2._get_where_condition(sql4)
    assert (
        sap2._get_where_condition(sql7)
        == "c = 1 AND d = 2 AND longcolname = 12345 AND otherlongcolname = 6789"
    ), sap2._get_where_condition(sql7)


def test__get_limit_v2():
    assert sap2._get_limit(sql1) is None
    assert sap2._get_limit(sql2) == 5
    assert sap2._get_limit(sql7) == 5


def test__get_offset_v2():
    assert sap2._get_offset(sql1) is None
    assert sap2._get_offset(sql2) == 3
    assert sap2._get_offset(sql7) == 10


def test_client_side_filters_simple_v2():
    _ = sap2._get_where_condition(sql5)
    assert sap2.client_side_filters == OrderedDict(
        {"AND": "longword123=5"}
    ), sap2.client_side_filters


def test_client_side_filters_with_limit_offset_v2():
    _ = sap2._get_where_condition(sql6)
    assert sap2.client_side_filters == OrderedDict(
        {"AND": "otherlongcolname=5"}
    ), sap2.client_side_filters

    _ = sap2._get_where_condition(sql7)
    assert sap2.client_side_filters == OrderedDict(
        {"AND": "thirdlongcolname = 01234"}
    ), sap2.client_side_filters


def test___build_pandas_filter_query_v2():
    _ = sap2._get_where_condition(sql6)
    assert (
        sap2._build_pandas_filter_query(sap2.client_side_filters)
        == "otherlongcolname == 5"
    ), sap2._build_pandas_filter_query(sap2.client_side_filters)
    _ = sap2._get_where_condition(sql7)
    assert (
        sap2._build_pandas_filter_query(sap2.client_side_filters)
        == "thirdlongcolname == 01234"
    ), sap2._build_pandas_filter_query(sap2.client_side_filters)



import json
import os
from unittest import mock

import pandas as pd
import pytest

from viadot.flows import HubspotToADLS

DATA = {
    "id": {"0": "820306930"},
    "createdAt": {"0": "2020-08-31T11:58:55.762Z"},
    "updatedAt": {"0": "2021-07-17T05:25:41.480Z"},
    "archived": {"0": "False"},
    "properties.amount": {"0": "3744.000"},
    "properties.createdate": {"0": "2020-08-31T11:58:55.762Z"},
    "properties.hs_lastmodifieddate": {"0": "2021-07-17T05:25:41.480Z"},
    "properties.hs_object_id": {"0": "820306930"},
    "properties.hs_product_id": {"0": "2344"},
    "properties.quantity": {"0": "2"},
}

ADLS_FILE_NAME = "test_hubspot.parquet"
ADLS_DIR_PATH = "raw/tests/"


@mock.patch(
    "viadot.tasks.HubspotToDF.run",
    return_value=pd.DataFrame(data=DATA),
)
@pytest.mark.run
def test_hubspot_to_adls_flow_run(mocked_class):
    flow = HubspotToADLS(
        "test_hubspot_to_adls_flow_run",
        hubspot_credentials_key="HUBSPOT",
        endpoint="line_items",
        filters=[
            {
                "filters": [
                    {
                        "propertyName": "createdate",
                        "operator": "BETWEEN",
                        "highValue": "2021-01-01",
                        "value": "2021-01-01",
                    },
                    {"propertyName": "quantity", "operator": "EQ", "value": "2"},
                ]
            },
            {
                "filters": [
                    {"propertyName": "amount", "operator": "EQ", "value": "3744.000"}
                ]
            },
        ],
        overwrite_adls=True,
        adls_dir_path=ADLS_DIR_PATH,
        adls_file_name=ADLS_FILE_NAME,
    )
    result = flow.run()
    assert result.is_successful()
    os.remove("test_hubspot_to_adls_flow_run.parquet")
    os.remove("test_hubspot_to_adls_flow_run.json")



import os
from datetime import timedelta

import pendulum
from azure.identity import EnvironmentCredential
from azure.keyvault.secrets import SecretClient
from prefect.tasks.secrets import PrefectSecret
from prefect.tasks.secrets.base import SecretBase
from prefect.utilities.tasks import defaults_from_attrs


def get_key_vault(
    credentials: str, secret_client_kwargs: dict, vault_name: str = None
) -> SecretClient:
    """
    Get Azure Key Vault.

    Args:
        credentials (str): Azure Key Vault credentials.
        secret_client_kwargs (dict): Keyword arguments to forward to the SecretClient.
        vault_name (str, optional): The name of the vault. Defaults to None

    Returns: Azure Key Vault
    """
    if not vault_name:
        vault_name = PrefectSecret("AZURE_DEFAULT_KEYVAULT").run()
    if credentials:
        # set credentials as env variables so that they're discoverable by EnvironmentCredential
        key_vault_credentials = credentials["KEY_VAULT"][vault_name]
        os.environ["AZURE_TENANT_ID"] = key_vault_credentials["AZURE_TENANT_ID"]
        os.environ["AZURE_CLIENT_ID"] = key_vault_credentials["AZURE_CLIENT_ID"]
        os.environ["AZURE_CLIENT_SECRET"] = key_vault_credentials["AZURE_CLIENT_SECRET"]
    else:
        try:
            # we can read the credentials automatically if user uses the default name
            credentials = PrefectSecret("AZURE_CREDENTIALS").run()
            key_vault_credentials = credentials["KEY_VAULT"][vault_name]
            os.environ["AZURE_TENANT_ID"] = key_vault_credentials["AZURE_TENANT_ID"]
            os.environ["AZURE_CLIENT_ID"] = key_vault_credentials["AZURE_CLIENT_ID"]
            os.environ["AZURE_CLIENT_SECRET"] = key_vault_credentials[
                "AZURE_CLIENT_SECRET"
            ]
        except ValueError as e:
            # go to step 3 (attempt to read from env)
            pass
    credentials = EnvironmentCredential(additionally_allowed_tenants=["*"])
    vault_url = f"https://{vault_name}.vault.azure.net"
    key_vault = SecretClient(
        vault_url=vault_url, credential=credentials, **secret_client_kwargs
    )
    return key_vault


class AzureKeyVaultSecret(SecretBase):
    """
    Task for retrieving secrets from an Azure Key Vault and returning it as a dictionary.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    For authentication, there are two options: you can set the `AZURE_CREDENTIALS` Prefect Secret
    containing your Azure Key Vault credentials which will be passed directly to `SecretClient`, or you
    can [configure your flow's runtime
    environment](https://docs.microsoft.com/en-us/dotnet/api/azure.identity.environmentcredential?view=azure-dotnet)
    for `EnvironmentCredential`.

    Args:
        - secret (str, optional): the name of the secret to retrieve
        - vault_name (str): the name of the vault from which to fetch the secret
        - secret_client_kwargs (dict, optional): additional keyword arguments to forward to the SecretClient.
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        secret: str = None,
        vault_name: str = None,
        secret_client_kwargs: dict = None,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        **kwargs,
    ):
        self.secret = secret
        self.vault_name = vault_name

        if secret_client_kwargs is None:
            self.secret_client_kwargs = {}
        else:
            self.secret_client_kwargs = secret_client_kwargs

        super().__init__(max_retries=max_retries, retry_delay=retry_delay, **kwargs)

    @defaults_from_attrs("secret", "vault_name")
    def run(
        self,
        secret: str = None,
        vault_name: str = None,
        credentials: dict = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> str:
        """
        Task run method.

        Args:
            - secret (str): the name of the secret to retrieve
            - vault_name (str): the name of the vault from which to fetch the secret
            - credentials (dict, optional): your Azure Key Vault credentials passed from an upstream
                Secret task. By default, credentials are read from the `AZURE_CREDENTIALS`
                Prefect Secret; this Secret must be a JSON string
                with the subkey `KEY_VAULT` and then vault_name containing three keys:
                `AZURE_TENANT_ID`, `AZURE_CLIENT_ID`, and `AZURE_CLIENT_SECRET`, which will be
                passed directly to `SecretClient`.  If not provided here or in context, the task
                will fall back to Azure credentials discovery using `EnvironmentCredential()`.
                Example `AZURE_CREDENTIALS` environment variable:
                `export AZURE_CREDENTIALS = '{"KEY_VAULT": {"test_key_vault": {"AZURE_TENANT_ID": "a", "AZURE_CLIENT_ID": "b", "AZURE_CLIENT_SECRET": "c"}}}'`

        Example:
            ```python
            from prefect import Flow
            from viadot.tasks import AzureKeyVaultSecret

            azure_secret_task = AzureKeyVaultSecret()

            with Flow(name="example") as f:
                secret = azure_secret_task(secret="test", vault_name="my_vault_name")
            out = f.run()
            ```
        Returns:
            - str: the contents of this secret, as a string
        """

        if secret is None:
            raise ValueError("A secret name must be provided.")

        key_vault = get_key_vault(
            vault_name=vault_name,
            credentials=credentials,
            secret_client_kwargs=self.secret_client_kwargs,
        )

        secret_string = key_vault.get_secret(secret).value

        return secret_string


class CreateAzureKeyVaultSecret(SecretBase):
    """
    Task for creating secrets in an Azure Key Vault.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    For authentication, there are two options: you can set the `AZURE_CREDENTIALS` Prefect Secret
    containing your Azure Key Vault credentials which will be passed directly to `SecretClient`, or you
    can [configure your flow's runtime
    environment](https://docs.microsoft.com/en-us/dotnet/api/azure.identity.environmentcredential?view=azure-dotnet)
    for `EnvironmentCredential`.

    Args:
        - secret (str, optional): the name of the secret to retrieve
        - vault_name (str): the name of the vault from which to fetch the secret
        - secret_client_kwargs (dict, optional): additional keyword arguments to forward to the SecretClient.
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        secret: str = None,
        value: str = None,
        lifetime: int = 365,
        vault_name: str = None,
        secret_client_kwargs: dict = None,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        **kwargs,
    ):
        self.secret = secret
        self.value = value
        self.lifetime = lifetime
        self.vault_name = vault_name

        if secret_client_kwargs is None:
            self.secret_client_kwargs = {}
        else:
            self.secret_client_kwargs = secret_client_kwargs

        super().__init__(max_retries=max_retries, retry_delay=retry_delay, **kwargs)

    @defaults_from_attrs(
        "secret",
        "value",
        "lifetime",
        "vault_name",
    )
    def run(
        self,
        secret: str = None,
        value: str = None,
        lifetime: int = None,
        vault_name: str = None,
        credentials: dict = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> bool:
        """
        Task run method.

        Args:
            - secret (str): the name of the secret to set
            - value (str): the value which the secret will hold
            - lifetime (int): The number of days after which the secret should expire.
            - vault_name (str): the name of the vault from which to fetch the secret
            - credentials (dict, optional): your Azure Key Vault credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with the subkey `KEY_VAULT` and then vault_name containing three keys:
                `AZURE_TENANT_ID`, `AZURE_CLIENT_ID`, and `AZURE_CLIENT_SECRET`, which will be
                passed directly to `SecretClient`.  If not provided here or in context, the task
                will fall back to Azure credentials discovery using `EnvironmentCredential()`.
                Example `AZURE_CREDENTIALS` environment variable:
                `export AZURE_CREDENTIALS = '{"KEY_VAULT": {"test_key_vault": {"AZURE_TENANT_ID": "a", "AZURE_CLIENT_ID": "b", "AZURE_CLIENT_SECRET": "c"}}}'`

        Example:
            ```python
            from prefect import Flow
            from prefect.tasks.secrets import PrefectSecret
            from viadot.tasks import CreateAzureKeyVaultSecret

            create_secret_task = CreateAzureKeyVaultSecret()

            with Flow(name="example") as f:
                azure_credentials = PrefectSecret("AZURE_CREDENTIALS")
                secret = create_secret_task(secret="test2", value=42, vault_name="my_vault_name", credentials=azure_credentials)
            out = f.run()
            ```
        Returns:
            - bool: Whether the secret was created successfully.
        """

        if secret is None:
            raise ValueError("A secret name must be provided.")

        key_vault = get_key_vault(
            vault_name=vault_name,
            credentials=credentials,
            secret_client_kwargs=self.secret_client_kwargs,
        )

        expires_on = pendulum.now("UTC").add(days=lifetime)
        secret_obj = key_vault.set_secret(secret, value, expires_on=expires_on)
        was_successful = secret_obj.name == secret

        return was_successful


class DeleteAzureKeyVaultSecret(SecretBase):
    """
    Task for removing ("soft delete") a secret from an Azure Key Vault.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    For authentication, there are two options: you can set the `AZURE_CREDENTIALS` Prefect Secret
    containing your Azure Key Vault credentials which will be passed directly to `SecretClient`, or you
    can [configure your flow's runtime
    environment](https://docs.microsoft.com/en-us/dotnet/api/azure.identity.environmentcredential?view=azure-dotnet)
    for `EnvironmentCredential`.

    Args:
        - secret (str, optional): the name of the secret to retrieve
        - vault_name (str): the name of the vault from which to fetch the secret
        - secret_client_kwargs (dict, optional): additional keyword arguments to forward to the SecretClient.
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        secret: str = None,
        vault_name: str = None,
        secret_client_kwargs: dict = None,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        **kwargs,
    ):
        self.secret = secret
        self.vault_name = vault_name

        if secret_client_kwargs is None:
            self.secret_client_kwargs = {}
        else:
            self.secret_client_kwargs = secret_client_kwargs

        super().__init__(max_retries=max_retries, retry_delay=retry_delay, **kwargs)

    @defaults_from_attrs("secret", "vault_name")
    def run(
        self,
        secret: str = None,
        vault_name: str = None,
        credentials: dict = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> bool:
        """
        Task run method.

        Args:
            - secret (str): the name of the secret to delete
            - vault_name (str): the name of the vault whethe the secret is located
            - credentials (dict, optional): your Azure Key Vault credentials passed from an upstream
                Secret task. By default, credentials are read from the `AZURE_CREDENTIALS`
                Prefect Secret; this Secret must be a JSON string
                with the subkey `KEY_VAULT` and then vault_name containing three keys:
                `AZURE_TENANT_ID`, `AZURE_CLIENT_ID`, and `AZURE_CLIENT_SECRET`, which will be
                passed directly to `SecretClient`.  If not provided here or in context, the task
                will fall back to Azure credentials discovery using `EnvironmentCredential()`.
                Example `AZURE_CREDENTIALS` environment variable:
                `export AZURE_CREDENTIALS = '{"KEY_VAULT": {"test_key_vault": {"AZURE_TENANT_ID": "a", "AZURE_CLIENT_ID": "b", "AZURE_CLIENT_SECRET": "c"}}}'`

        Example:
            ```python
            from prefect import Flow
            from viadot.tasks import DeleteAzureKeyVaultSecret

            azure_secret_task = DeleteAzureKeyVaultSecret()

            with Flow(name="example") as f:
                secret = azure_secret_task(secret="test", vault_name="my_vault_name")
            out = f.run()
            ```
        Returns:
            - bool: Whether the secret was deleted successfully.
        """

        if secret is None:
            raise ValueError("A secret name must be provided.")

        key_vault = get_key_vault(
            vault_name=vault_name,
            credentials=credentials,
            secret_client_kwargs=self.secret_client_kwargs,
        )

        poller = key_vault.begin_delete_secret(secret)
        poller.wait(timeout=60 * 5)
        was_successful = poller.status() == "finished"

        return was_successful



import json
from datetime import timedelta
from typing import Any, Dict, List

import pandas as pd
from prefect import Task
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities.tasks import defaults_from_attrs

from ..sources import Salesforce
from .azure_key_vault import AzureKeyVaultSecret


def get_credentials(credentials_secret: str, vault_name: str = None):
    """
    Get Salesforce credentials from Azure Key Vault.

    Args:
        credentials_secret (str): The name of the Azure Key Vault secret containing a dictionary with
            the required credentials (eg. username, password, token). Defaults to None.
        vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.

    Returns: Credentials
    """
    if not credentials_secret:
        # attempt to read a default for the service principal secret name
        try:
            credentials_secret = PrefectSecret("SALESFORCE_DEFAULT_SECRET").run()
        except ValueError:
            pass

    if credentials_secret:
        azure_secret_task = AzureKeyVaultSecret()
        credentials_str = azure_secret_task.run(
            secret=credentials_secret, vault_name=vault_name
        )
        credentials = json.loads(credentials_str)

        return credentials


class SalesforceUpsert(Task):
    """
    Task for upserting a pandas DataFrame to Salesforce.

    Args:
    """

    def __init__(
        self,
        table: str = None,
        external_id: str = None,
        domain: str = "test",
        client_id: str = "viadot",
        env: str = "DEV",
        raise_on_error: bool = False,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.table = table
        self.external_id = external_id
        self.domain = domain
        self.client_id = client_id
        self.env = env
        self.raise_on_error = raise_on_error

        super().__init__(
            name="salesforce_upsert",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Upserting data to Salesforce"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "table",
        "external_id",
        "domain",
        "client_id",
        "env",
        "raise_on_error",
        "max_retries",
        "retry_delay",
    )
    def run(
        self,
        df: pd.DataFrame = None,
        table: str = None,
        external_id: str = None,
        domain: str = None,
        client_id: str = None,
        credentials_secret: str = None,
        vault_name: str = None,
        env: str = None,
        raise_on_error: bool = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> None:
        """Task run method.

        Args:
            df (pd.DataFrame, optional): The DataFrame to upsert. Defaults to None.
            table (str, optional): The table where the data should be upserted. Defaults to None.
            external_id (str, optional): The external ID to use for the upsert. Defaults to None.
            domain (str, optional): Domain of a connection; defaults to 'test' (sandbox).
                Can only be added if built-in username/password/security token is provided. Defaults to None.
            client_id (str, optional): Client id to keep the track of API calls. Defaults to None.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
                the required credentials (eg. username, password, token). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
            env (str, optional): Environment information, provides information about credential and connection configuration. Defaults to 'DEV'.
            raise_on_error (bool, optional): Whether to raise an exception if a row upsert fails.
                If False, we only display a warning. Defaults to False.
        """
        credentials = get_credentials(credentials_secret, vault_name=vault_name)
        salesforce = Salesforce(
            credentials=credentials,
            env=env,
            domain=domain,
            client_id=client_id,
        )
        self.logger.info(f"Upserting {df.shape[0]} rows to Salesforce...")
        salesforce.upsert(
            df=df, table=table, external_id=external_id, raise_on_error=raise_on_error
        )
        self.logger.info(f"Successfully upserted {df.shape[0]} rows to Salesforce.")


class SalesforceBulkUpsert(Task):
    """
    Task for upserting a pandas DataFrame to Salesforce.

    Args:
    """

    def __init__(
        self,
        table: str = None,
        external_id: str = None,
        domain: str = "test",
        client_id: str = "viadot",
        env: str = "DEV",
        raise_on_error: bool = False,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.table = table
        self.external_id = external_id
        self.domain = domain
        self.client_id = client_id
        self.env = env
        self.raise_on_error = raise_on_error

        super().__init__(
            name="salesforce_bulk_upsert",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Upserting data to Salesforce"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "table",
        "external_id",
        "domain",
        "client_id",
        "env",
        "raise_on_error",
        "max_retries",
        "retry_delay",
    )
    def run(
        self,
        df: pd.DataFrame = None,
        table: str = None,
        external_id: str = None,
        batch_size: int = None,
        domain: str = None,
        client_id: str = None,
        credentials_secret: str = None,
        vault_name: str = None,
        env: str = None,
        raise_on_error: bool = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> None:
        """Task run method.

        Args:
            df (pd.DataFrame, optional): The DataFrame to upsert. Defaults to None.
            table (str, optional): The table where the data should be upserted. Defaults to None.
            external_id (str, optional): The external ID to use for the upsert. Defaults to None.
            domain (str, optional): Domain of a connection; defaults to 'test' (sandbox).
                Can only be added if built-in username/password/security token is provided. Defaults to None.
            client_id (str, optional): Client id to keep the track of API calls. Defaults to None.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
                the required credentials (eg. username, password, token). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
            env (str, optional): Environment information, provides information about credential and connection configuration. Defaults to 'DEV'.
            raise_on_error (bool, optional): Whether to raise an exception if a row upsert fails.
                If False, we only display a warning. Defaults to False.
        """
        credentials = get_credentials(credentials_secret, vault_name=vault_name)
        salesforce = Salesforce(
            credentials=credentials,
            env=env,
            domain=domain,
            client_id=client_id,
        )
        self.logger.info(f"Upserting {df.shape[0]} rows to Salesforce...")
        salesforce.bulk_upsert(
            df=df,
            table=table,
            external_id=external_id,
            batch_size=batch_size,
            raise_on_error=raise_on_error,
        )
        self.logger.info(f"Successfully upserted {df.shape[0]} rows to Salesforce.")


class SalesforceToDF(Task):
    """
    The task for querying Salesforce and saving data as the data frame.

    Args:
        query (str, optional): Query for download the data if specific download is needed. Defaults to None.
        table (str, optional): Table name. Can be used instead of query. Defaults to None.
        columns (List[str], optional): List of columns which are needed - table argument is needed. Defaults to None.
        domain (str, optional): Domain of a connection. defaults to 'test' (sandbox).
            Can only be added if built-in username/password/security token is provided. Defaults to None.
        client_id (str, optional): Client id to keep the track of API calls. Defaults to None.
        env (str, optional): Environment information, provides information about credential
            and connection configuration. Defaults to 'DEV'.
    """

    def __init__(
        self,
        query: str = None,
        table: str = None,
        columns: List[str] = None,
        domain: str = "test",
        client_id: str = "viadot",
        env: str = "DEV",
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        self.query = query
        self.table = table
        self.columns = columns
        self.domain = domain
        self.client_id = client_id
        self.env = env

        super().__init__(
            name="salesforce_to_df",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self):
        """Download Salesforce data to a DF"""
        super().__call__(self)

    @defaults_from_attrs(
        "query",
        "table",
        "columns",
        "domain",
        "client_id",
        "env",
    )
    def run(
        self,
        query: str = None,
        table: str = None,
        columns: List[str] = None,
        env: str = None,
        domain: str = None,
        client_id: str = None,
        credentials_secret: str = None,
        vault_name: str = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        """
        Task run method.

        Args:
            query (str, optional): Query for download the data if specific download is needed. Defaults to None.
            table (str, optional): Table name. Can be used instead of query. Defaults to None.
            columns (List[str], optional): List of columns which are needed - table argument is needed. Defaults to None.
            env (str, optional): Environment information, provides information about credential
                and connection configuration. Defaults to 'DEV'.
            domain (str, optional): Domain of a connection. defaults to 'test' (sandbox).
                Can only be added if built-in username/password/security token is provided. Defaults to None.
            client_id (str, optional): Client id to keep the track of API calls. Defaults to None.
            credentials_secret (str, optional): The name of the Azure Key Vault secret for Salesforce. Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
        """
        credentials = get_credentials(credentials_secret, vault_name=vault_name)
        salesforce = Salesforce(
            credentials=credentials,
            env=env,
            domain=domain,
            client_id=client_id,
        )
        self.logger.info(f"Retreiving the data from Salesforce...")
        df = salesforce.to_df(
            query=query, table=table, columns=columns, if_empty="replace"
        )
        self.logger.info(f"Successfully downloaded data from Salesforce.")

        return df



import xml.etree.ElementTree as ET
from typing import Any, Dict, List, Optional
from xml.etree.ElementTree import fromstring

import pandas as pd
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

from ..sources import BusinessCore


class BusinessCoreToParquet(Task):
    def __init__(
        self,
        path: str,
        url: str,
        filters_dict: Dict[str, Any] = {
            "BucketCount": None,
            "BucketNo": None,
            "FromDate": None,
            "ToDate": None,
        },
        verify: bool = True,
        credentials: Dict[str, Any] = None,
        config_key: str = "BusinessCore",
        if_empty: str = "skip",
        timeout=3600,
        *args,
        **kwargs,
    ):
        """Task for downloading  data from Business Core API to a Parquet file.

        Args:
            path (str, required): Path where to save a Parquet file.
            url (str, required): Base url to a view in Business Core API.
            filters_dict (Dict[str, Any], optional): Filters in form of dictionary. Available filters: 'BucketCount',
                'BucketNo', 'FromDate', 'ToDate'.  Defaults to {"BucketCount": None,"BucketNo": None,"FromDate": None,
                "ToDate": None,}.
            verify (bool, optional): Whether or not verify certificates while connecting to an API. Defaults to True.
            credentials (Dict[str, Any], optional): Credentials stored in a dictionary. Required credentials: username,
                password. Defaults to None.
            config_key (str, optional): Credential key to dictionary where details are stored. Defaults to "BusinessCore".
            if_empty (str, optional): What to do if output DataFrame is empty. Defaults to "skip".
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """

        self.url = url
        self.path = path
        self.credentials = credentials
        self.config_key = config_key
        self.filters_dict = filters_dict
        self.verify = verify
        self.if_empty = if_empty

        super().__init__(
            name="business_core_to_parquet",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Load Business Core data to Parquet"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "url", "path", "credentials", "config_key", "filters_dict", "verify", "if_empty"
    )
    def run(
        self,
        path: str = None,
        url: str = None,
        credentials: Dict[str, Any] = None,
        config_key: str = None,
        filters_dict: str = None,
        verify: bool = True,
        if_empty: str = None,
    ):
        """Run method for BusinessCoreToParquet task. Saves data from Business Core API to Parquet file.

        Args:
            path (str, optional): Path where to save a Parquet file. Defaults to None.
            url (str, optional): Base url to a view in Business Core API. Defaults to None.
            filters_dict (Dict[str, Any], optional): Filters in form of dictionary. Available filters: 'BucketCount',
                'BucketNo', 'FromDate', 'ToDate'.  Defaults to None.
            verify (bool, optional): Whether or not verify certificates while connecting to an API. Defaults to True.
            credentials (Dict[str, Any], optional): Credentials stored in a dictionary. Required credentials: username,
                password. Defaults to None.
            config_key (str, optional): Credential key to dictionary where details are stored. Defaults to None.
            if_empty (str, optional): What to do if output DataFrame is empty. Defaults to None.


        """
        bc = BusinessCore(
            url=url,
            credentials=credentials,
            config_key=config_key,
            filters_dict=filters_dict,
            verify=verify,
        )
        return bc.to_parquet(path=path, if_empty=if_empty)



import pytest

from viadot.flows import AzureSQLTransform
from viadot.tasks import AzureSQLDBQuery

SCHEMA = "sandbox"
TABLE = "test"
FQN = f"{SCHEMA}.{TABLE}"


@pytest.fixture()
def TEST_TABLE():
    run_sql_task = AzureSQLDBQuery()
    run_sql_task.run(
        f""" 
            DROP TABLE IF EXISTS {FQN};
            CREATE TABLE {FQN} (id INT, name VARCHAR(25))"""
    )
    run_sql_task.run(f"INSERT INTO {FQN} VALUES (1, 'Mike')")
    yield
    run_sql_task.run(f"DROP TABLE {FQN}")


def test_azure_sql_transform(TEST_TABLE):
    flow = AzureSQLTransform(
        name="test",
        query=f"SELECT * FROM {FQN}",
    )
    result = flow.run()
    task = list(result.result.keys())[0]
    actual_result = result.result[task].result[0]
    assert list(actual_result) == [1, "Mike"]



from typing import List

import pandas as pd
import pandas_gbq
import pytest
from numpy import ndarray

from viadot.exceptions import CredentialError, DBDataAccessError
from viadot.sources import BigQuery


@pytest.fixture(scope="function")
def BIGQ():
    """
    Fixture for creating a BigQuery class instance. This fixture initializes a BigQuery client
    using the provided credentials key and yields the class instance.
    The class instance can be used within a test function to interact with BigQuery.

    Yields:
        BigQuery: A BigQuery client instance.
    """
    BQ = BigQuery(credentials_key="BIGQUERY_TESTS")
    yield BQ


@pytest.fixture(scope="session")
def insert_into_tables() -> None:
    """
    A function to insert data into a BigQuery table. In the current version, tables are deleted
    after 60 days. This operation is used to secure tests and structure in a BigQuery project.
    """
    table_id1 = "manigeo.manigeo_tab"
    table_id2 = "manigeo.space"
    df = pd.DataFrame({"my_value": ["val1", "val2", "val3"]})
    pandas_gbq.to_gbq(df, table_id1, if_exists="replace")
    pandas_gbq.to_gbq(df, table_id2, if_exists="replace")


# SQL query for public dataset - user with access to Bigquery can also use public datasets and tables.
QUERY = """
        SELECT name, SUM(number) AS total
        FROM `bigquery-public-data.usa_names.usa_1910_2013`
        GROUP BY name, gender
        ORDER BY total DESC 
        LIMIT 4
        """


def test_credentials():
    """Test to see if an exception is thrown if credentials are not provided."""
    with pytest.raises(CredentialError, match=r"Credentials not found."):
        BigQuery(credentials_key="BIGQUERY_TESTS_FAKE")


def test_list_project(BIGQ):
    """
    Testing the correctness of the project name.

    Args:
        BIGQ (Bigquery): Bigquery class instance.
    """
    project = BIGQ.get_project_id()

    assert project == "manifest-geode-341308"


def test_list_datasets(BIGQ):
    """
    Testing the correctness of dataset names.

    Args:
        BIGQ (Bigquery): Bigquery class instance.
    """
    datasets = list(BIGQ.list_datasets())

    assert datasets == ["manigeo", "official_empty"]


def test_list_tables(BIGQ):
    """
    Testing the correctness of table names.

    Args:
        BIGQ (Bigquery): Bigquery class instance.
    """
    datasets = BIGQ.list_datasets()
    tables = list(BIGQ.list_tables(datasets[0]))
    assert "space" and "manigeo_tab" in tables


def test_list_columns(BIGQ):
    """
    Testing the validity of a column name in a specific table in BigQuery and the return type.

    Args:
        BIGQ (Bigquery): Bigquery class instance.
    """
    columns = BIGQ.list_columns(dataset_name="manigeo", table_name="space")

    assert "my_value" in columns
    assert isinstance(columns, ndarray)


def test_query_is_df(BIGQ):
    """
    Testing the return type of `query_to_df` function. It should be a Data Frame.

    Args:
        BIGQ (Bigquery): Bigquery class instance.
    """
    df = BIGQ.query_to_df(QUERY)

    assert isinstance(df, pd.DataFrame)


def test_query(BIGQ):
    """
    Testing the corectness of `query_to_df`execution.

    Args:
        BIGQ (Bigquery): Bigquery class instance.
    """
    df = BIGQ.query_to_df(QUERY)
    total_received = df["total"].values

    assert total_received == [4924235, 4818746, 4703680, 4280040]


def test_wrong_query(BIGQ):
    """
    Testing if the exception is raised with invalid query.

    Args:
        BIGQ (Bigquery): Bigquery class instance.
    """
    fake_query = """
            SELECT fake_name
            FROM `bigquery-public-data.usa_names.fake_table`
            ORDER BY fake_name DESC 
            LIMIT 4
            """
    with pytest.raises(DBDataAccessError):
        BIGQ.query_to_df(fake_query)



import json
import os
from os.path import dirname, expanduser, join, realpath

import pytest

from viadot.config import Config

USER_HOME = expanduser("~")
UNIT_TESTS_DIR_PATH = dirname(realpath(__file__))
UNIT_TESTS_CONFIG_PATH = join(UNIT_TESTS_DIR_PATH, "credentials.json")
DEFAULT_CONFIG_PATH = join(USER_HOME, ".config", "credentials.json")


@pytest.fixture(scope="session")
def config():
    CONFIG_PATH = os.environ.get("VIADOT_CONFIG_PATH") or DEFAULT_CONFIG_PATH
    yield CONFIG_PATH
    os.rename(CONFIG_PATH + "_renamed", CONFIG_PATH)


def get_credentials(key: str):
    with open(UNIT_TESTS_CONFIG_PATH, "r") as f:
        credentials = json.load(f)
    return credentials[key]


def test_credentials():
    test_credentials = Config.from_json(UNIT_TESTS_CONFIG_PATH, key="TEST_STANZA")
    credentials = get_credentials("TEST_STANZA")
    assert test_credentials == credentials


def test_no_credentials_file(config):
    os.rename(config, config + "_renamed")
    from viadot.sources import UKCarbonIntensity

    UKCarbonIntensity()



from datetime import datetime
from typing import List, Literal

import prefect
from prefect import Task
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

logger = logging.get_logger()


def iter_throught_flow_runs(flow_runs_details: List[dict] = None) -> dict:
    """
    Generate Flow run details from dict containing flow runs

    Args:
        flow_runs_details (List[dict], optional): List of Flow run details. Defaults to None.

    Yields:
        dict: Flow run details
    """
    for x in range(len(flow_runs_details)):
        for flow_run in flow_runs_details[x]["flow_runs"]:
            yield flow_run


def get_time_from_last_successful_run(flow_runs_details: List[dict] = None) -> str:
    """
    Get start_time from last flow run where state was success.

    Args:
        flow_runs_details (List[dict], optional): List of Flow run details. Defaults to None.

    Returns:
        str: Flow run start_time
    """

    for flow_run in iter_throught_flow_runs(flow_runs_details=flow_runs_details):
        if flow_run["state"] == "Success":
            return flow_run["start_time"]


def calculate_difference(
    date_to_compare: str = None,
    base_date: str = str(datetime.today()),
    diff_type: Literal["time", "date"] = "date",
):
    """
    Calculate diffrence between two dates.

    Args:
        date_to_compare (str, optional): Date to compare with base_date (Flow run start_time). Defaults to None.
        base_date (str, optional): The base date (be saved as Prefect schedule date. Defaults to str(datetime.today()).
        diff_type (Literal["time", "date"], optional): _description_. Defaults to "date".

     Returns:
        (int, float): Differences in days when calculating date or time (hours, minutes) when calculating time.
    """
    base_date = get_formatted_date(base_date, diff_type)
    date_to_compare = get_formatted_date(date_to_compare, diff_type)

    if diff_type == "date":
        difference = abs(base_date - date_to_compare)
        return difference.days

    if diff_type == "time":
        difference_h = abs(base_date.hour - date_to_compare.hour)
        difference_m = date_to_compare.minute - base_date.minute
        if difference_h == 1:
            if difference_m < 0:
                return 0
            if difference_m > 0:
                return float(f"1.{(abs(difference_m))}")
            if difference_m == 0:
                return 1
        if difference_h < 1:
            return 0
        if difference_h > 1:
            return difference_h


def check_if_scheduled_run(time_run: str = None, time_schedule: str = None) -> bool:
    """
    Check if run was scheduled or started by user.

    Args:
        time_run (str, optional): The time the Flow was started. Defaults to None.
        time_schedule (str, optional): Scheduled time of Flow. Defaults to None.

    Returns:
        bool: True if flow run was started automatically. False if Flow was started by user.
    """
    diff = calculate_difference(
        date_to_compare=time_run,
        base_date=time_schedule,
        diff_type="time",
    )
    if diff <= 1:
        return True
    if diff > 1:
        return False


def get_formatted_date(
    time_unclean: str = None,
    return_value: Literal["time", "date"] = "date",
):
    """
    Format date from "2022-02-21T01:00:00+00:00" to date or time.

    Args:
        time_unclean (str, optional): Time in datetime format obtained from Prefect. Defaults to None.
        return_value (Literal["time", "date"], optional): Choose the format to be extracted from datetime - time or date.
                                                        Defaults to "date".

    Returns:
        datetime: Date (datetime.date) or time (datetime.time)
    """
    if return_value == "time":
        time_extracted = time_unclean.split("T")[1]
        time_clean_str = time_extracted.split(".")[0]
        time_clean = datetime.strptime(time_clean_str[:8], "%H:%M:%S")
        return time_clean.time()

    if return_value == "date":
        date_extracted = time_unclean.split("T")[0]
        date_clean = datetime.strptime(date_extracted, "%Y-%m-%d")
        return date_clean.date()


class GetFlowNewDateRange(Task):
    """
    Get the flow date_range_type and add the days if the last run of the flow was unsuccessful.
    A task that checks the status of the last flow run and change it if the last flow has failed.

    If it is originally 'last_14_days' but the flow has run in Failed state for the last 2 days
    then from 'last_14_days' it will be 'last_16_days' in the next flow_run.

    Note that here is the check whether the flow run has been planned.
    Only If the flow was run by a prefect (according to the planned schedule), a change of date dates is possible.
    """

    def __init__(
        self,
        flow_name: str = None,
        date_range_type: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """
        Initialize GetFlowNewDateRange class.

        Args:
            flow_name (str, optional): Prefect flow name. Defaults to None.
            date_range_type (str, optional): String argument that should look like this: 'last_X_days' -
                X is a number of days. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """

        self.flow_name = flow_name
        self.date_range_type = date_range_type

        super().__init__(
            name="prefect_extract_details",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self):
        """Extract time from Prefect Flow run"""
        super().__call__(self)

    def change_date_range(self, date_range: str = None, difference: int = None) -> str:
        """
        Replace number of days in string "last_X_days" based on the sum of number extracted from string
        and difference passed to function.

        Args:
            date_range (str, optional): Date range in format "last_X_days". Defaults to None.
            difference (int, optional): Int value representing days. Defaults to None.

        Returns:
            str: Date range in "last_X_days" format
        """
        old_range_splitted = date_range.split("_")
        old_range = int(old_range_splitted[1])
        new_range = old_range + difference

        new_range_splitted = old_range_splitted
        new_range_splitted[1] = str(new_range)
        date_range_type = "_".join(new_range_splitted)
        return date_range_type

    @defaults_from_attrs(
        "flow_name",
        "date_range_type",
    )
    def run(
        self,
        flow_name,
        date_range_type,
        **kwargs,
    ) -> None:
        query = (
            """
             {           
                flow (where: { name: { _eq: "%s" } } )
                {
                flow_runs(
                    order_by: {end_time: desc}
                    where: {start_time:{ _is_null:false } } ) 
                    {
                      id
                      end_time
                      start_time
                      state
                      scheduled_start_time
                    }  
                } 
            }
        """
            % flow_name
        )

        client = prefect.Client()
        flow_runs = client.graphql(query)
        flow_runs_details = flow_runs.data.flow

        for flow_run in iter_throught_flow_runs(flow_runs_details=flow_runs_details):
            if flow_run["scheduled_start_time"]:
                time_schedule = flow_run["scheduled_start_time"]
                break
        last_success_start_time = get_time_from_last_successful_run(flow_runs_details)
        is_scheduled = check_if_scheduled_run(
            time_run=last_success_start_time,
            time_schedule=time_schedule,
        )

        if is_scheduled is True:
            difference_days = calculate_difference(
                date_to_compare=last_success_start_time,
                base_date=time_schedule,
                diff_type="date",
            )
            new_date_range_type = self.change_date_range(
                date_range=date_range_type, difference=difference_days
            )
            return new_date_range_type

        if is_scheduled is False:
            return date_range_type



import os
from pathlib import Path
from typing import Any, Dict, List

import pendulum
from prefect import Flow, task
from prefect.backend import set_key_value
from prefect.utilities import logging

logger = logging.get_logger()

from viadot.task_utils import (
    add_ingestion_metadata_task,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_csv,
    df_to_parquet,
    dtypes_to_json_task,
)
from viadot.tasks import AzureDataLakeUpload
from viadot.tasks.sharepoint import SharepointToDF


class SharepointToADLS(Flow):
    def __init__(
        self,
        name: str,
        url_to_file: str = None,
        nrows_to_df: int = None,
        path_to_file: str = None,
        sheet_number: int = None,
        validate_excel_file: bool = False,
        output_file_extension: str = ".csv",
        local_dir_path: str = None,
        adls_dir_path: str = None,
        adls_file_name: str = None,
        adls_sp_credentials_secret: str = None,
        overwrite_adls: bool = False,
        if_empty: str = "warn",
        if_exists: str = "replace",
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading Excel file from Sharepoint then uploading it to Azure Data Lake.

        Args:
            name (str): The name of the flow.
            url_to_file (str, optional): Link to a file on Sharepoint. Defaults to None.
                        (e.g : https://{tenant_name}.sharepoint.com/sites/{folder}/Shared%20Documents/Dashboard/file).
            nrows_to_df (int, optional): Number of rows to read at a time. Defaults to 50000. Defaults to None.
            path_to_file (str, optional): Path to local Excel file. Defaults to None.
            sheet_number (int, optional): Sheet number to be extracted from file. Counting from 0, if None all sheets are axtracted. Defaults to None.
            validate_excel_file (bool, optional): Check if columns in separate sheets are the same. Defaults to False.
            output_file_extension (str, optional): Output file extension - to allow selection of .csv for data which is not easy to handle with parquet. Defaults to ".csv".
            local_dir_path (str, optional): File directory. Defaults to None.
            adls_dir_path (str, optional): Azure Data Lake destination folder/catalog path. Defaults to None.
            adls_file_name (str, optional): Name of file in ADLS. Defaults to None.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite files in the lake. Defaults to False.
            if_empty (str, optional): What to do if query returns no data. Defaults to "warn".
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        # SharepointToDF
        self.if_empty = if_empty
        self.nrows = nrows_to_df
        self.path_to_file = path_to_file
        self.url_to_file = url_to_file
        self.local_dir_path = local_dir_path
        self.sheet_number = sheet_number
        self.validate_excel_file = validate_excel_file
        self.timeout = timeout

        # AzureDataLakeUpload
        self.overwrite = overwrite_adls
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.if_exists = if_exists
        self.output_file_extension = output_file_extension
        self.now = str(pendulum.now("utc"))
        if self.local_dir_path is not None:
            self.local_file_path = (
                self.local_dir_path + self.slugify(name) + self.output_file_extension
            )
        else:
            self.local_file_path = self.slugify(name) + self.output_file_extension
        self.local_json_path = self.slugify(name) + ".json"
        self.adls_dir_path = adls_dir_path
        if adls_file_name is not None:
            self.adls_file_path = os.path.join(adls_dir_path, adls_file_name)
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", Path(adls_file_name).stem + ".json"
            )
        else:
            self.adls_file_path = os.path.join(
                adls_dir_path, self.now + self.output_file_extension
            )
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", self.now + ".json"
            )

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    def gen_flow(self) -> Flow:
        excel_to_df_task = SharepointToDF(timeout=self.timeout)
        df = excel_to_df_task.bind(
            path_to_file=self.path_to_file,
            url_to_file=self.url_to_file,
            nrows=self.nrows,
            sheet_number=self.sheet_number,
            validate_excel_file=self.validate_excel_file,
            flow=self,
        )

        df_with_metadata = add_ingestion_metadata_task.bind(df, flow=self)
        dtypes_dict = df_get_data_types_task.bind(df_with_metadata, flow=self)
        df_mapped = df_map_mixed_dtypes_for_parquet.bind(
            df_with_metadata, dtypes_dict, flow=self
        )
        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df_mapped,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        dtypes_to_json_task.bind(
            dtypes_dict=dtypes_dict, local_json_path=self.local_json_path, flow=self
        )
        json_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        json_to_adls_task.bind(
            from_path=self.local_json_path,
            to_path=self.adls_schema_file_dir_file,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        df_mapped.set_upstream(df_with_metadata, flow=self)
        dtypes_to_json_task.set_upstream(df_mapped, flow=self)
        df_to_file.set_upstream(dtypes_to_json_task, flow=self)

        file_to_adls_task.set_upstream(df_to_file, flow=self)
        json_to_adls_task.set_upstream(dtypes_to_json_task, flow=self)
        set_key_value(key=self.adls_dir_path, value=self.adls_file_path)

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()



import os
from typing import Any, Dict, List, Literal

import prefect
from prefect import Flow, task
from prefect.utilities import logging

from viadot.task_utils import df_to_csv as df_to_csv_task
from viadot.task_utils import get_sql_dtypes_from_df as get_sql_dtypes_from_df_task
from viadot.tasks import BCPTask, DuckDBQuery, DuckDBToDF, SQLServerCreateTable

logger = logging.get_logger(__name__)


@task(timeout=3600)
def cleanup_csv_task(path: str):
    logger = prefect.context.get("logger")

    logger.info(f"Removing file {path}...")
    try:
        os.remove(path)
        logger.info(f"File {path} has been successfully removed.")
        return True
    except Exception as e:
        logger.exception(f"File {path} could not be removed.")
        return False


class DuckDBToSQLServer(Flow):
    def __init__(
        self,
        name: str,
        duckdb_schema: str = None,
        duckdb_table: str = None,
        if_empty: str = "warn",
        duckdb_query: str = None,
        duckdb_credentials: dict = None,
        local_file_path: str = None,
        write_sep: str = "\t",
        sql_server_schema: str = None,
        sql_server_table: str = None,
        dtypes: Dict[str, Any] = None,
        if_exists: Literal["fail", "replace", "append", "delete"] = "replace",
        sql_server_credentials: dict = None,
        on_bcp_error: Literal["skip", "fail"] = "skip",
        bcp_error_log_path="./log_file.log",
        tags: List[str] = ["load"],
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for moving a table from DuckDB to SQL Server. User can specify schema and table to transfer
        data between databases or use a SELECT query instead.

        Args:
            name (str): The name of the flow.
            duckdb_schema (str, optional): Destination schema. Defaults to None.
            duckdb_table (str, optional): Destination table. Defaults to None.
            if_empty (str, optional): What to do if the query returns no data. Defaults to "warn".
            duckdb_query (str, optional): SELECT query to be executed on DuckDB, its result will be used to
            create table. Defaults to None.
            duckdb_credentials (dict, optional): The config to use for connecting with DuckDB.
            local_file_path (str, optional): Local destination path. Defaults to None.
            write_sep (str, optional): The delimiter for the output CSV file. Defaults to "\t".
            sql_server_schema (str, optional): Destination schema. Defaults to None.
            sql_server_table (str, optional): Destination table. Defaults to None.
            dtypes (dict, optional): The data types to be enforced for the resulting table. By default,
            we infer them from the DataFrame. Defaults to None.
            if_exists (Literal, optional): What to do if the table exists. Defaults to "replace".
            sql_server_credentials (dict, optional): The credentials to use for connecting with SQL Server.
            on_bcp_error (Literal["skip", "fail"], optional): What to do if error occurs. Defaults to "skip".
            bcp_error_log_path (string, optional): Full path of an error file. Defaults to "./log_file.log".
            tags (List[str], optional): Flow tags to use, eg. to control flow concurrency. Defaults to ["load"].
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """

        # DuckDBToDF
        self.duckdb_schema = duckdb_schema
        self.duckdb_table = duckdb_table
        self.duckdb_credentials = duckdb_credentials

        # DuckDBQuery
        self.duckdb_query = duckdb_query

        # df_to_csv_task
        self.local_file_path = local_file_path or self.slugify(name) + ".csv"
        self.write_sep = write_sep
        self.if_empty = if_empty

        # SQLServerCreateTable
        self.sql_server_table = sql_server_table
        self.sql_server_schema = sql_server_schema
        self.dtypes = dtypes
        self.if_exists = self._map_if_exists(if_exists)
        self.sql_server_credentials = sql_server_credentials
        self.on_bcp_error = on_bcp_error
        self.bcp_error_log_path = bcp_error_log_path

        # Global
        self.tags = tags
        self.timeout = timeout

        super().__init__(*args, name=name, **kwargs)

        self.duckdb_run_query_task = DuckDBQuery(credentials=duckdb_credentials)

        self.gen_flow()

    @staticmethod
    def _map_if_exists(if_exists: str) -> str:
        mapping = {"append": "skip"}
        return mapping.get(if_exists, if_exists)

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        if self.duckdb_query is None:
            duckdb_to_df_task = DuckDBToDF(timeout=self.timeout)
            df = duckdb_to_df_task.bind(
                schema=self.duckdb_schema,
                table=self.duckdb_table,
                if_empty=self.if_empty,
                credentials=self.duckdb_credentials,
                flow=self,
            )
        else:
            df = self.duckdb_run_query_task.bind(
                query=self.duckdb_query,
                credentials=self.duckdb_credentials,
                fetch_type="dataframe",
                flow=self,
            )

        df_to_csv = df_to_csv_task.bind(
            df=df,
            path=self.local_file_path,
            sep=self.write_sep,
            flow=self,
        )
        if self.dtypes:
            # Use user-provided dtypes.
            dtypes = self.dtypes
        else:
            dtypes = get_sql_dtypes_from_df_task.bind(df=df, flow=self)

        create_table_task = SQLServerCreateTable(timeout=self.timeout)
        create_table_task.bind(
            schema=self.sql_server_schema,
            table=self.sql_server_table,
            dtypes=dtypes,
            if_exists=self.if_exists,
            credentials=self.sql_server_credentials,
            flow=self,
        )

        bulk_insert_task = BCPTask(timeout=self.timeout)
        bulk_insert_task.bind(
            path=self.local_file_path,
            schema=self.sql_server_schema,
            table=self.sql_server_table,
            credentials=self.sql_server_credentials,
            on_error=self.on_bcp_error,
            error_log_file_path=self.bcp_error_log_path,
            flow=self,
        )
        cleanup_csv_task.bind(path=self.local_file_path, flow=self)

        create_table_task.set_upstream(df_to_csv, flow=self)
        bulk_insert_task.set_upstream(create_table_task, flow=self)
        cleanup_csv_task.set_upstream(bulk_insert_task, flow=self)



import os

import pandas as pd
import pytest
from prefect.tasks.secrets import PrefectSecret

from viadot.config import local_config
from viadot.exceptions import CredentialError
from viadot.sources import Sharepoint
from viadot.task_utils import df_get_data_types_task
from viadot.tasks.sharepoint import SharepointToDF


def get_url() -> str:
    """
    Function to get file URL.

    Returns:
        str: File URL.
    """
    return local_config["SHAREPOINT"].get("url")


@pytest.fixture(scope="session")
def sharepoint():
    """
    Fixture for creating a Sharepoint class instance.
    The class instance can be used within a test functions to interact with Sharepoint.
    """
    s = Sharepoint()
    yield s


@pytest.fixture(scope="session")
def file_name(sharepoint):
    """
    A function built to get the path to a file.

    Args:
        sharepoint (Sharepoint): Sharepoint class instance.
    """
    path = "Questionnaires.xlsx"
    sharepoint.download_file(download_to_path=path, download_from_path=get_url())
    yield path
    os.remove(path)


def test_credentials_not_found():
    """
    Testing if a VauleError is thrown when none of credentials are given.

    Args:
        sharepoint (Sharepoint): Sharepoint class instance.
    """
    none_credentials = None
    with pytest.raises(CredentialError, match=r"Credentials not found."):
        Sharepoint(credentials=none_credentials)


def test_get_connection_credentials():
    """
    Testing if a CredentialError is thrown when credentials doesn't contain required keys.

    Args:
        sharepoint (Sharepoint): Sharepoint class instance.
    """
    credentials = {"site": "tenant.sharepoint.com", "username": "User"}
    s = Sharepoint(credentials=credentials)
    with pytest.raises(CredentialError, match="Missing credentials."):
        s.get_connection()


def test_connection(sharepoint):
    """
    Testing if connection is succesfull with given credentials.

    Args:
        sharepoint (Sharepoint): Sharepoint class instance.
    """
    credentials = local_config.get("SHAREPOINT")
    site = f'https://{credentials["site"]}'
    conn = sharepoint.get_connection()
    response = conn.get(site)
    assert response.status_code == 200


def test_sharepoint_to_df_task():
    """Testing if result of `SharepointToDF` is a Data Frame."""
    task = SharepointToDF()
    credentials_secret = PrefectSecret("SHAREPOINT_KV").run()
    res = task.run(
        credentials_secret=credentials_secret,
        sheet_number=0,
        path_to_file="Questionnaires.xlsx",
        url_to_file=get_url(),
    )
    assert isinstance(res, pd.DataFrame)
    os.remove("Questionnaires.xlsx")


def test_download_file_missing_patameters(sharepoint):
    """
    Testing if a VauleError is thrown when none of the parameters are given.

    Args:
        sharepoint (Sharepoint): Sharepoint class instance.
    """
    with pytest.raises(ValueError, match=r"Missing required parameter"):
        sharepoint.download_file(download_to_path=None, download_from_path=None)


def test_file_download(file_name):
    """
    Testing if file is downloaded.

    Args:
        file_name (str): File name.
    """
    files = []
    for file in os.listdir():
        if os.path.isfile(os.path.join(file)):
            files.append(file)
    assert file_name in files


def test_autopopulating_download_from(file_name):
    """
    Testing if file name is correct.

    Args:
        file_name (str): File name.
    """
    assert os.path.basename(get_url()) == file_name


def test_file_extension():
    """Testing if file has correct extension."""
    file_ext = (".xlsm", ".xlsx")
    assert get_url().endswith(file_ext)


def test_file_to_df(file_name):
    """
    Testing if downloaded file contains data and if first sheet can be build as a Data frame.

    Args:
        file_name (str): File name.
    """
    df = pd.read_excel(file_name, sheet_name=0)
    df_test = pd.DataFrame(data={"col1": [1, 2]})
    assert type(df) == type(df_test)


def test_get_data_types(file_name):
    """
    Testing if downloaded file contains data and columns have `String` type.

    Args:
        file_name (str): File name.
    """
    df = pd.read_excel(file_name, sheet_name=0)
    dtypes_map = df_get_data_types_task.run(df)
    dtypes = dtypes_map.values()

    assert "String" in dtypes



from typing import Any, Dict, List, Union

from prefect import Flow, Task, apply_map
from prefect.utilities import logging

from ..tasks import BlobFromCSV, CreateTableFromBlob, SupermetricsToCSV

logger = logging.get_logger(__name__)


class SupermetricsToAzureSQL(Flow):
    def __init__(
        self,
        name: str,
        ds_id: str,
        ds_accounts: List[str],
        ds_user: str,
        fields: List[str],
        ds_segments: List[str] = None,
        date_range_type: str = None,
        start_date: str = None,
        end_date: str = None,
        settings: Dict[str, Any] = None,
        filter: str = None,
        max_rows: int = 1000000,
        max_columns: int = None,
        order_columns: str = None,
        dtypes: Dict[str, Any] = None,
        blob_path: str = None,
        overwrite_blob: bool = True,
        table: str = None,
        schema: str = None,
        local_file_path: str = None,
        if_exists: str = "replace",  # this applies to the full CSV file, not per chunk
        if_empty: str = "warn",
        max_download_retries: int = 5,
        supermetrics_task_timeout: int = 60 * 30,
        parallel: bool = True,
        tags: List[str] = ["extract"],
        sep: str = "\t",
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        self.ds_id = ds_id
        self.ds_accounts = ds_accounts
        self.ds_segments = ds_segments
        self.ds_user = ds_user
        self.fields = fields
        self.date_range_type = date_range_type
        self.start_date = start_date
        self.end_date = end_date
        self.settings = settings
        self.filter = filter
        self.max_rows = max_rows
        self.max_columns = max_columns
        self.order_columns = order_columns
        self.local_file_path = local_file_path or self.slugify(name) + ".csv"
        self.blob_path = blob_path
        self.overwrite_blob = overwrite_blob
        self.table = table
        self.schema = schema
        self.dtypes = dtypes
        self.if_exists = if_exists
        self.if_empty = if_empty
        self.max_download_retries = max_download_retries
        self.supermetrics_task_timeout = supermetrics_task_timeout
        self.parallel = parallel
        self.tags = tags
        self.sep = sep
        self.timeout = timeout
        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_supermetrics_task(
        self, ds_accounts: Union[str, List[str]], flow: Flow = None
    ) -> Task:
        supermetrics_to_csv_task = SupermetricsToCSV(timeout=self.timeout)
        t = supermetrics_to_csv_task.bind(
            ds_id=self.ds_id,
            ds_accounts=ds_accounts,
            ds_segments=self.ds_segments,
            ds_user=self.ds_user,
            fields=self.fields,
            date_range_type=self.date_range_type,
            start_date=self.start_date,
            end_date=self.end_date,
            settings=self.settings,
            filter=self.filter,
            max_rows=self.max_rows,
            max_columns=self.max_columns,
            order_columns=self.order_columns,
            path=self.local_file_path,
            if_exists="append",
            if_empty=self.if_empty,
            max_retries=self.max_download_retries,
            timeout=self.supermetrics_task_timeout,
            sep=self.sep,
            flow=flow,
        )
        return t

    def gen_flow(self) -> Flow:
        if self.parallel:
            # generate a separate task for each account
            supermetrics_downloads = apply_map(
                self.gen_supermetrics_task, self.ds_accounts, flow=self
            )
        else:
            supermetrics_downloads = self.gen_supermetrics_task(
                ds_accounts=self.ds_accounts, flow=self
            )

        csv_to_blob_storage_task = BlobFromCSV(timeout=self.timeout)
        csv_to_blob_storage_task.bind(
            from_path=self.local_file_path,
            to_path=self.blob_path,
            overwrite=self.overwrite_blob,
            flow=self,
        )
        blob_to_azure_sql_task = CreateTableFromBlob(timeout=self.timeout)
        blob_to_azure_sql_task.bind(
            blob_path=self.blob_path,
            schema=self.schema,
            table=self.table,
            dtypes=self.dtypes,
            sep=self.sep,
            if_exists=self.if_exists,
            flow=self,
        )

        csv_to_blob_storage_task.set_upstream(supermetrics_downloads, flow=self)
        blob_to_azure_sql_task.set_upstream(csv_to_blob_storage_task, flow=self)



import pandas as pd
import pytest

from viadot.sources import Salesforce


@pytest.fixture(scope="session")
def salesforce():
    s = Salesforce()
    yield s


@pytest.fixture(scope="session")
def test_df_external():
    data = {
        "Id": ["111"],
        "LastName": ["John Tester-External"],
        "SAPContactId__c": ["112"],
    }
    df = pd.DataFrame(data=data)
    yield df


def test_upsert_empty(salesforce):
    try:
        df = pd.DataFrame()
        salesforce.upsert(df=df, table="Contact")
    except Exception as exception:
        assert False, exception


def test_upsert_external_id_correct(salesforce, test_df_external):
    try:
        salesforce.upsert(
            df=test_df_external, table="Contact", external_id="SAPContactId__c"
        )
    except Exception as exception:
        assert False, exception
    result = salesforce.download(table="Contact")
    exists = list(
        filter(lambda contact: contact["LastName"] == "John Tester-External", result)
    )
    assert exists != None


def test_upsert_external_id_wrong(salesforce, test_df_external):
    with pytest.raises(ValueError):
        salesforce.upsert(df=test_df_external, table="Contact", external_id="SAPId")


def test_download_no_query(salesforce):
    ordered_dict = salesforce.download(table="Account")
    assert len(ordered_dict) > 0


def test_download_with_query(salesforce):
    query = "SELECT Id, Name FROM Account"
    ordered_dict = salesforce.download(query=query)
    assert len(ordered_dict) > 0


def test_to_df(salesforce):
    df = salesforce.to_df(table="Account")
    assert df.empty == False



from typing import Any, Dict

from azure.storage.blob import BlobClient, BlobServiceClient

from ..config import local_config
from .base import Source


class AzureBlobStorage(Source):
    """
    A class for pulling data from the Azure Blob Storage
    (do not confuse with Azure Data Lake Gen1 and Azure Data Lake Gen2).

    Documentation for this API is located at: https://supermetrics.com/docs/product-api-getting-started/

    Parameters
    ----------
    credentials : Dict[str, Any], optional
        Credentials containing a connection string, by default None
    """

    def __init__(self, credentials: Dict[str, Any] = None, *args, **kwargs):
        DEFAULT_CREDENTIALS = local_config.get("AZURE_BLOB_STORAGE")
        credentials = credentials or DEFAULT_CREDENTIALS
        super().__init__(*args, credentials=credentials, **kwargs)

    def to_storage(self, from_path: str, to_path: str, overwrite: bool = False):
        """Upload a file to the storage.

        Args:
            from_path (str): Path to the local file to be uploaded to the storge.
            to_path (str): The destination path in the format 'container/path/a.csv'
            overwrite (bool, optional): [description]. Defaults to False.

        Example:
        ```python
        from viadot.sources import AzureBlobStorage
        storage = AzureBlobStorage()
        storage.to_storage('tests/test.csv')
        ```

        Returns:
            bool: Whether the operation was successful.
        """
        conn_str = self.credentials["CONNECTION_STRING"]
        blob_service_client = BlobServiceClient.from_connection_string(conn_str)

        container_name = to_path.split("/")[0]
        blob_path = "/".join(to_path.split("/")[1:])
        blob_client = blob_service_client.get_blob_client(
            container=container_name, blob=blob_path
        )

        with open(from_path, "rb") as data:
            blob_client.upload_blob(data, overwrite=overwrite)

        return True

    def exists(self, path: str) -> bool:
        """
        Check if blob exists in Azure Storage.

        Args:
            path (str): The bolb path in the format 'container/path/a.csv'

        Example:
        ```python
        from viadot.sources.azure_blob_storage import AzureBlobStorage

        path = "tests/test.csv"
        container_name = path.split("/")[0]
        blob_path = "/".join(path.split("/")[1:])
        conn_str = self.credentials["CONNECTION_STRING"]
        blob = BlobClient.from_connection_string(
            conn_str=conn_str, container_name=container_name, blob_name=blob_path
        )

        blob.exists()
        ```

        Returns:
            bool: Whether the operation was successful.
        """
        container_name = path.split("/")[0]
        blob_path = "/".join(path.split("/")[1:])
        conn_str = self.credentials["CONNECTION_STRING"]

        blob = BlobClient.from_connection_string(
            conn_str=conn_str, container_name=container_name, blob_name=blob_path
        )
        if_exists = blob.exists()
        return if_exists



import json
import os
from datetime import timedelta
from typing import List

import numpy as np
import pandas as pd
from prefect import Task
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities.tasks import defaults_from_attrs

from ..sources import AzureDataLake
from .azure_key_vault import AzureKeyVaultSecret


class AzureDataLakeDownload(Task):
    """
    Task for downloading data from the Azure Data lakes (gen1 and gen2).

    Args:
        from_path (str, optional): The path from which to download the file(s). Defaults to None.
        to_path (str, optional): The destination path. Defaults to None.
        recursive (bool, optional): Set this to true if downloading entire directories.
        gen (int, optional): The generation of the Azure Data Lake. Defaults to 2.
        vault_name (str, optional): The name of the vault from which to fetch the secret. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
        max_retries (int, optional): [description]. Defaults to 3.
        retry_delay (timedelta, optional): [description]. Defaults to timedelta(seconds=10).
    """

    def __init__(
        self,
        from_path: str = None,
        to_path: str = None,
        recursive: bool = False,
        gen: int = 2,
        vault_name: str = None,
        timeout: int = 3600,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        *args,
        **kwargs,
    ):
        self.from_path = from_path
        self.to_path = to_path
        self.recursive = recursive
        self.gen = gen
        self.vault_name = vault_name

        super().__init__(
            name="adls_download",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Download file(s) from the Azure Data Lake"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "from_path",
        "to_path",
        "recursive",
        "gen",
        "vault_name",
        "max_retries",
        "retry_delay",
    )
    def run(
        self,
        from_path: str = None,
        to_path: str = None,
        recursive: bool = None,
        gen: int = None,
        sp_credentials_secret: str = None,
        vault_name: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> None:
        """Task run method.

        Args:
            from_path (str): The path from which to download the file(s).
            to_path (str): The destination path.
            recursive (bool): Set this to true if downloading entire directories.
            gen (int): The generation of the Azure Data Lake.
            sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """
        file_name = from_path.split("/")[-1]
        to_path = to_path or file_name

        if not sp_credentials_secret:
            # attempt to read a default for the service principal secret name
            try:
                sp_credentials_secret = PrefectSecret(
                    "AZURE_DEFAULT_ADLS_SERVICE_PRINCIPAL_SECRET"
                ).run()
            except ValueError:
                pass

        if sp_credentials_secret:
            azure_secret_task = AzureKeyVaultSecret()
            credentials_str = azure_secret_task.run(
                secret=sp_credentials_secret, vault_name=vault_name
            )
            credentials = json.loads(credentials_str)
        else:
            credentials = {
                "ACCOUNT_NAME": os.environ["AZURE_ACCOUNT_NAME"],
                "AZURE_TENANT_ID": os.environ["AZURE_TENANT_ID"],
                "AZURE_CLIENT_ID": os.environ["AZURE_CLIENT_ID"],
                "AZURE_CLIENT_SECRET": os.environ["AZURE_CLIENT_SECRET"],
            }
        lake = AzureDataLake(gen=gen, credentials=credentials)

        full_dl_path = os.path.join(credentials["ACCOUNT_NAME"], from_path)
        self.logger.info(f"Downloading data from {full_dl_path} to {to_path}...")
        lake.download(from_path=from_path, to_path=to_path, recursive=recursive)
        self.logger.info(f"Successfully downloaded data to {to_path}.")


class AzureDataLakeUpload(Task):
    """Upload file(s) to Azure Data Lake.

    Args:
        from_path (str, optional): The local path from which to upload the file(s). Defaults to None.
        to_path (str, optional): The destination path. Defaults to None.
        recursive (bool, optional): Set this to true if uploading entire directories. Defaults to False.
        overwrite (bool, optional): Whether to overwrite files in the lake. Defaults to False.
        gen (int, optional): The generation of the Azure Data Lake. Defaults to 2.
        vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        from_path: str = None,
        to_path: str = None,
        recursive: bool = False,
        overwrite: bool = False,
        gen: int = 2,
        vault_name: str = None,
        timeout: int = 3600,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        *args,
        **kwargs,
    ):
        self.from_path = from_path
        self.to_path = to_path
        self.recursive = recursive
        self.overwrite = overwrite
        self.gen = gen
        self.vault_name = vault_name

        super().__init__(
            name="adls_upload",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Upload file(s) to the Azure Data Lake"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "from_path",
        "to_path",
        "recursive",
        "overwrite",
        "gen",
        "vault_name",
        "max_retries",
        "retry_delay",
    )
    def run(
        self,
        from_path: str = None,
        to_path: str = None,
        recursive: bool = None,
        overwrite: bool = None,
        gen: int = None,
        sp_credentials_secret: str = None,
        vault_name: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> None:
        """Task run method.

        Args:
            from_path (str): The path from which to upload the file(s).
            to_path (str): The destination path.
            recursive (bool): Set to true if uploading entire directories.
            overwrite (bool): Whether to overwrite the file(s) if they exist.
            gen (int): The generation of the Azure Data Lake.
            sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """

        if not sp_credentials_secret:
            # attempt to read a default for the service principal secret name
            try:
                sp_credentials_secret = PrefectSecret(
                    "AZURE_DEFAULT_ADLS_SERVICE_PRINCIPAL_SECRET"
                ).run()
            except ValueError:
                pass

        if sp_credentials_secret:
            azure_secret_task = AzureKeyVaultSecret()
            credentials_str = azure_secret_task.run(
                secret=sp_credentials_secret, vault_name=vault_name
            )
            credentials = json.loads(credentials_str)
        else:
            credentials = {
                "ACCOUNT_NAME": os.environ["AZURE_ACCOUNT_NAME"],
                "AZURE_TENANT_ID": os.environ["AZURE_TENANT_ID"],
                "AZURE_CLIENT_ID": os.environ["AZURE_CLIENT_ID"],
                "AZURE_CLIENT_SECRET": os.environ["AZURE_CLIENT_SECRET"],
            }
        lake = AzureDataLake(gen=gen, credentials=credentials)

        full_to_path = os.path.join(credentials["ACCOUNT_NAME"], to_path)
        self.logger.info(f"Uploading data from {from_path} to {full_to_path}...")
        lake.upload(
            from_path=from_path,
            to_path=to_path,
            recursive=recursive,
            overwrite=overwrite,
        )
        self.logger.info(f"Successfully uploaded data to {full_to_path}.")


class AzureDataLakeToDF(Task):
    def __init__(
        self,
        path: str = None,
        sep: str = "\t",
        quoting: int = 0,
        lineterminator: str = None,
        error_bad_lines: bool = None,
        gen: int = 2,
        vault_name: str = None,
        timeout: int = 3600,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        *args,
        **kwargs,
    ):
        """Load file(s) from the Azure Data Lake to a pandas DataFrame.
        Currently supports CSV and parquet files.

        Args:
            path (str, optional): The path from which to load the DataFrame. Defaults to None.
            sep (str, optional): The separator to use when reading a CSV file. Defaults to "\t".
            quoting (int, optional): The quoting mode to use when reading a CSV file. Defaults to 0.
            lineterminator (str, optional): The newline separator to use when reading a CSV file. Defaults to None.
            error_bad_lines (bool, optional): Whether to raise an exception on bad lines. Defaults to None.
            gen (int, optional): The generation of the Azure Data Lake. Defaults to 2.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        self.path = path
        self.sep = sep
        self.quoting = quoting
        self.lineterminator = lineterminator
        self.error_bad_lines = error_bad_lines

        self.gen = gen
        self.vault_name = vault_name
        super().__init__(
            name="adls_to_df",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Load file(s) from the Azure Data Lake to a pandas DataFrame."""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "path",
        "sep",
        "quoting",
        "lineterminator",
        "error_bad_lines",
        "gen",
        "vault_name",
        "max_retries",
        "retry_delay",
    )
    def run(
        self,
        path: str = None,
        sep: str = None,
        quoting: int = None,
        lineterminator: str = None,
        error_bad_lines: bool = None,
        gen: int = None,
        sp_credentials_secret: str = None,
        vault_name: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> pd.DataFrame:
        """Task run method.

        Args:
            path (str): The path to file(s) which should be loaded into a DataFrame.
            sep (str): The field separator to use when loading the file to the DataFrame.
            quoting (int, optional): The quoting mode to use when reading a CSV file. Defaults to 0.
            lineterminator (str, optional): The newline separator to use when reading a CSV file. Defaults to None.
            error_bad_lines (bool, optional): Whether to raise an exception on bad lines. Defaults to None.
            gen (int): The generation of the Azure Data Lake.
            sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """

        if quoting is None:
            quoting = 0

        if path is None:
            raise ValueError("Please provide the path to the file to be downloaded.")

        if not sp_credentials_secret:
            # attempt to read a default for the service principal secret name
            try:
                sp_credentials_secret = PrefectSecret(
                    "AZURE_DEFAULT_ADLS_SERVICE_PRINCIPAL_SECRET"
                ).run()
            except ValueError:
                pass

        if sp_credentials_secret:
            azure_secret_task = AzureKeyVaultSecret()
            credentials_str = azure_secret_task.run(
                secret=sp_credentials_secret, vault_name=vault_name
            )
            credentials = json.loads(credentials_str)
        else:
            credentials = {
                "ACCOUNT_NAME": os.environ["AZURE_ACCOUNT_NAME"],
                "AZURE_TENANT_ID": os.environ["AZURE_TENANT_ID"],
                "AZURE_CLIENT_ID": os.environ["AZURE_CLIENT_ID"],
                "AZURE_CLIENT_SECRET": os.environ["AZURE_CLIENT_SECRET"],
            }
        lake = AzureDataLake(gen=gen, credentials=credentials, path=path)

        full_dl_path = os.path.join(credentials["ACCOUNT_NAME"], path)
        self.logger.info(f"Downloading data from {full_dl_path} to a DataFrame...")
        df = lake.to_df(
            sep=sep,
            quoting=quoting,
            lineterminator=lineterminator,
            error_bad_lines=error_bad_lines,
        )
        self.logger.info(f"Successfully loaded data.")
        return df


class AzureDataLakeCopy(Task):
    """
    Task for copying data between the Azure Data lakes files.

    Args:
        from_path (str, optional): The path from which to copy the file(s). Defaults to None.
        to_path (str, optional): The destination path. Defaults to None.
        recursive (bool, optional): Set this to true if copy entire directories.
        gen (int, optional): The generation of the Azure Data Lake. Defaults to 2.
        vault_name (str, optional): The name of the vault from which to fetch the secret. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
        max_retries (int, optional): [description]. Defaults to 3.
        retry_delay (timedelta, optional): [description]. Defaults to timedelta(seconds=10).
    """

    def __init__(
        self,
        from_path: str = None,
        to_path: str = None,
        recursive: bool = False,
        gen: int = 2,
        vault_name: str = None,
        timeout: int = 3600,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        *args,
        **kwargs,
    ):
        self.from_path = from_path
        self.to_path = to_path
        self.recursive = recursive
        self.gen = gen
        self.vault_name = vault_name

        super().__init__(
            name="adls_copy",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Copy file(s) from the Azure Data Lake"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "from_path",
        "to_path",
        "recursive",
        "gen",
        "vault_name",
        "max_retries",
        "retry_delay",
    )
    def run(
        self,
        from_path: str = None,
        to_path: str = None,
        recursive: bool = None,
        gen: int = None,
        sp_credentials_secret: str = None,
        vault_name: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> None:
        """Task run method.

        Args:
            from_path (str): The path from which to copy the file(s).
            to_path (str): The destination path.
            recursive (bool): Set this to true if copying entire directories.
            gen (int): The generation of the Azure Data Lake.
            sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """
        file_name = from_path.split("/")[-1]
        to_path = to_path or file_name

        if not sp_credentials_secret:
            # attempt to read a default for the service principal secret name
            try:
                sp_credentials_secret = PrefectSecret(
                    "AZURE_DEFAULT_ADLS_SERVICE_PRINCIPAL_SECRET"
                ).run()
            except ValueError:
                pass

        if sp_credentials_secret:
            azure_secret_task = AzureKeyVaultSecret()
            credentials_str = azure_secret_task.run(
                secret=sp_credentials_secret, vault_name=vault_name
            )
            credentials = json.loads(credentials_str)
        else:
            credentials = {
                "ACCOUNT_NAME": os.environ["AZURE_ACCOUNT_NAME"],
                "AZURE_TENANT_ID": os.environ["AZURE_TENANT_ID"],
                "AZURE_CLIENT_ID": os.environ["AZURE_CLIENT_ID"],
                "AZURE_CLIENT_SECRET": os.environ["AZURE_CLIENT_SECRET"],
            }
        lake = AzureDataLake(gen=gen, credentials=credentials)

        full_dl_path = os.path.join(credentials["ACCOUNT_NAME"], from_path)
        self.logger.info(f"Copying data from {full_dl_path} to {to_path}...")
        lake.cp(from_path=from_path, to_path=to_path, recursive=recursive)
        self.logger.info(f"Successfully copied data to {to_path}.")


class AzureDataLakeList(Task):
    """
    Task for listing files in Azure Data Lake.

    Args:
        path (str, optional): The path to the directory which contents you want to list. Defaults to None.
        gen (int, optional): The generation of the Azure Data Lake. Defaults to 2.
        vault_name (str, optional): The name of the vault from which to fetch the secret. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
        max_retries (int, optional): [description]. Defaults to 3.
        retry_delay (timedelta, optional): [description]. Defaults to timedelta(seconds=10).

    Returns:
        List[str]: The list of paths to the contents of `path`. These paths
        do not include the container, eg. the path to the file located at
        "https://my_storage_acc.blob.core.windows.net/raw/supermetrics/test_file.txt"
        will be shown as "raw/supermetrics/test_file.txt".
    """

    def __init__(
        self,
        path: str = None,
        gen: int = 2,
        vault_name: str = None,
        timeout: int = 3600,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        *args,
        **kwargs,
    ):
        self.path = path
        self.gen = gen
        self.vault_name = vault_name

        super().__init__(
            name="adls_list",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    @defaults_from_attrs(
        "path",
        "gen",
        "vault_name",
        "max_retries",
        "retry_delay",
    )
    def run(
        self,
        path: str = None,
        recursive: bool = False,
        file_to_match: str = None,
        gen: int = None,
        sp_credentials_secret: str = None,
        vault_name: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> List[str]:
        """Task run method.

        Args:
            path (str, optional): The path to the directory which contents you want to list. Defaults to None.
            recursive (bool, optional): If True, recursively list all subdirectories and files. Defaults to False.
            file_to_match (str, optional): If exist it only returns files with that name. Defaults to None.
            gen (int): The generation of the Azure Data Lake. Defaults to None.
            sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.

        Returns:
            List[str]: The list of paths to the contents of `path`. These paths
            do not include the container, eg. the path to the file located at
            "https://my_storage_acc.blob.core.windows.net/raw/supermetrics/test_file.txt"
            will be shown as "raw/supermetrics/test_file.txt".
        """

        if not sp_credentials_secret:
            # attempt to read a default for the service principal secret name
            try:
                sp_credentials_secret = PrefectSecret(
                    "AZURE_DEFAULT_ADLS_SERVICE_PRINCIPAL_SECRET"
                ).run()
            except ValueError:
                pass

        if sp_credentials_secret:
            azure_secret_task = AzureKeyVaultSecret()
            credentials_str = azure_secret_task.run(
                secret=sp_credentials_secret, vault_name=vault_name
            )
            credentials = json.loads(credentials_str)
        else:
            credentials = {
                "ACCOUNT_NAME": os.environ["AZURE_ACCOUNT_NAME"],
                "AZURE_TENANT_ID": os.environ["AZURE_TENANT_ID"],
                "AZURE_CLIENT_ID": os.environ["AZURE_CLIENT_ID"],
                "AZURE_CLIENT_SECRET": os.environ["AZURE_CLIENT_SECRET"],
            }
        lake = AzureDataLake(gen=gen, credentials=credentials)

        full_dl_path = os.path.join(credentials["ACCOUNT_NAME"], path)

        self.logger.info(f"Listing files in {full_dl_path}.")
        if recursive:
            self.logger.info("Loading ADLS directories recursively.")
            files = lake.find(path)
            if file_to_match:
                conditions = [file_to_match in item for item in files]
                valid_files = np.array([])
                if any(conditions):
                    index = np.where(conditions)[0]
                    files = list(np.append(valid_files, [files[i] for i in index]))
                else:
                    raise FileExistsError(
                        f"There are not any available file named {file_to_match}."
                    )
        else:
            files = lake.ls(path)

        self.logger.info(f"Successfully listed files in {full_dl_path}.")
        return files


class AzureDataLakeRemove(Task):
    """
    Task for removing objects from Azure Data Lake.

    Args:
        path (str, optional): The path to the directory from which to delete files. Defaults to None.
        recursive (bool): Set this to true if removing entire directories.
        gen (int, optional): The generation of the Azure Data Lake. Defaults to 2.
        vault_name (str, optional): The name of the vault from which to fetch the secret. Defaults to None.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
        max_retries (int, optional): Maximum number of retries before failing. Defaults to 3.
        retry_delay (timedelta, optional): Time to wait before the next retry attempt. Defaults to timedelta(seconds=10).
    """

    def __init__(
        self,
        path: str = None,
        recursive: bool = False,
        gen: int = 2,
        vault_name: str = None,
        timeout: int = 3600,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        *args,
        **kwargs,
    ):
        self.path = path
        self.recursive = recursive
        self.gen = gen
        self.vault_name = vault_name

        super().__init__(
            name="adls_rm",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    @defaults_from_attrs(
        "path",
        "recursive",
        "gen",
        "vault_name",
        "max_retries",
        "retry_delay",
    )
    def run(
        self,
        path: str = None,
        recursive: bool = None,
        gen: int = None,
        sp_credentials_secret: str = None,
        vault_name: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> List[str]:
        """Task run method.

        Args:
            path (str): The path to the directory or file you want to delete. Defaults to None.
            recursive (bool): Set this to True if removing files recursively. Defaults to False.
            gen (int): The generation of the Azure Data Lake. Defaults to None.
            sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
        """

        if not sp_credentials_secret:
            # attempt to read a default for the service principal secret name
            try:
                sp_credentials_secret = PrefectSecret(
                    "AZURE_DEFAULT_ADLS_SERVICE_PRINCIPAL_SECRET"
                ).run()
            except ValueError:
                pass

        if sp_credentials_secret:
            azure_secret_task = AzureKeyVaultSecret()
            credentials_str = azure_secret_task.run(
                secret=sp_credentials_secret, vault_name=vault_name
            )
            credentials = json.loads(credentials_str)
        else:
            credentials = {
                "ACCOUNT_NAME": os.environ["AZURE_ACCOUNT_NAME"],
                "AZURE_TENANT_ID": os.environ["AZURE_TENANT_ID"],
                "AZURE_CLIENT_ID": os.environ["AZURE_CLIENT_ID"],
                "AZURE_CLIENT_SECRET": os.environ["AZURE_CLIENT_SECRET"],
            }
        lake = AzureDataLake(gen=gen, credentials=credentials)
        full_path = os.path.join(credentials["ACCOUNT_NAME"], path)

        if full_path.endswith(".csv") or full_path.endswith(".parquet"):
            logger_details = "file from"
        else:
            logger_details = "directory"

        self.logger.info(f"Deleting {logger_details} {full_path}...")
        lake.rm(path, recursive=recursive)
        self.logger.info(f"Successfully deleted {logger_details} {full_path}.")



import os

import openpyxl
import pytest
from openpyxl import load_workbook

from viadot.tasks.open_apis.uk_carbon_intensity import StatsToCSV, StatsToExcel

TEST_FILE_PATH = "/home/viadot/tests/uk_carbon_intensity_test.csv"
TEST_FILE_PATH_EXCEL = "/home/viadot/tests/uk_carbon_intensity_test.xlsx"


@pytest.fixture(scope="session")
def ukci_task():
    ukci_task = StatsToCSV()
    yield ukci_task
    os.remove(TEST_FILE_PATH)


@pytest.fixture(scope="session")
def ukci_task_excel():
    ukci_task_excel = StatsToExcel()
    yield ukci_task_excel
    os.remove(TEST_FILE_PATH_EXCEL)


# def test_uk_carbon_intensity_to_csv(ukci_task):
#     ukci_task.run(path=TEST_FILE_PATH)
#     if_exist = os.path.isfile(TEST_FILE_PATH)
#     assert if_exist == True


# def test_uk_carbon_intensity_to_excel(ukci_task_excel):
#     ukci_task_excel.run(path=TEST_FILE_PATH_EXCEL)
#     if_exist = os.path.isfile(TEST_FILE_PATH_EXCEL)
#     assert if_exist == True


# def test_uk_carbon_intensity_to_excel_contain(ukci_task_excel):
#     ukci_task_excel.run(path=TEST_FILE_PATH_EXCEL)
#     excel_file = load_workbook(TEST_FILE_PATH_EXCEL)
#     value = excel_file["A1"].value
#     assert value == "from"



import os
import time
from typing import Any, Dict, List

import numpy as np
import pandas as pd
import prefect
from pandas import DataFrame
from prefect import Task
from prefect.engine import signals
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from viadot.exceptions import APIError
from viadot.sources import Genesys

logger = logging.get_logger()


class GenesysToCSV(Task):
    def __init__(
        self,
        report_name: str = "genesys_to_csv",
        view_type: str = "queue_performance_detail_view",
        post_data_list: List[str] = None,
        start_date: str = None,
        end_date: str = None,
        environment: str = None,
        report_url: str = None,
        report_columns: List[str] = None,
        local_file_path: str = "",
        sep: str = "\t",
        conversationId_list: List[str] = None,
        key_list: List[str] = None,
        credentials_genesys: Dict[str, Any] = None,
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """Task for downloading data from Genesys API to CSV.

        Args:
            report_name (str, optional): The name of this task. Defaults to a general name 'genesys_to_csv'.
            view_type (str, optional): The type of view export job to be created. Defaults to "queue_performance_detail_view".
            post_data_list (List[str], optional): List of string templates to generate json body. Defaults to None.
            credentials_genesys (Dict[str, Any], optional): Credentials to connect with Genesys API containing CLIENT_ID. Defaults to None.
            start_date (str, optional): Start date of the report. Defaults to None.
            end_date (str, optional): End date of the report. Defaults to None.
            environment (str, optional): Adress of host server. Defaults to None than will be used enviroment
            from credentials.
            report_url (str, optional): The url of report generated in json response. Defaults to None.
            report_columns (List[str], optional): List of exisiting column in report. Defaults to None.
            local_file_path (str, optional): The local path from which to upload the file(s). Defaults to "".
            sep (str, optional): Separator in csv file. Defaults to "\t".
            conversationId_list (List[str], optional): List of conversationId passed as attribute of GET method. Defaults to None.
            key_list (List[str], optional): List of keys needed to specify the columns in the GET request method. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """

        self.logger = prefect.context.get("logger")
        self.report_name = report_name
        self.view_type = view_type
        self.environment = environment
        self.report_url = report_url
        self.report_columns = report_columns
        self.post_data_list = post_data_list
        self.start_date = start_date
        self.end_date = end_date
        self.credentials_genesys = credentials_genesys
        self.local_file_path = local_file_path
        self.sep = sep
        self.conversationId_list = conversationId_list
        self.key_list = key_list

        super().__init__(
            name=self.report_name,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Download Genesys data to CSV"""
        return super().__call__(*args, **kwargs)

    def merge_conversations_dfs(self, data_to_merge: list) -> DataFrame:
        """Method to merge all the conversations data into a single data frame.

        Args:
            data_to_merge (list): List with all the conversations in json format.
            Example for all levels data to merge:
                {
                "conversations": [
                    {
                        **** LEVEL 0 data ****
                        "participants": [
                            {
                                **** LEVEL 1 data ****
                                "sessions": [
                                    {
                                        "agentBullseyeRing": 1,
                                        **** LEVEL 2 data ****
                                        "mediaEndpointStats": [
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                        ],
                                        "metrics": [
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                        ],
                                        "segments": [
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                        ],
                                    }
                                ],
                            },
                            {
                                "participantId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                **** LEVEL 1 data ****
                                "sessions": [
                                    {
                                        **** LEVEL 2 data ****
                                        "mediaEndpointStats": [
                                            {
                                                **** LEVEL 3 data ****
                                            }
                                        ],
                                        "flow": {
                                            **** LEVEL 2 data ****
                                        },
                                        "metrics": [
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                        ],
                                        "segments": [
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                        ],
                                    }
                                ],
                            },
                        ],
                    }
                ],
                "totalHits": 100,
            }

        Returns:
            DataFrame: A single data frame with all the content.
        """
        # LEVEL 0
        df0 = pd.json_normalize(data_to_merge)
        df0.drop(["participants"], axis=1, inplace=True)

        # LEVEL 1
        df1 = pd.json_normalize(
            data_to_merge,
            record_path=["participants"],
            meta=["conversationId"],
        )
        df1.drop(["sessions"], axis=1, inplace=True)

        # LEVEL 2
        df2 = pd.json_normalize(
            data_to_merge,
            record_path=["participants", "sessions"],
            meta=[
                ["participants", "externalContactId"],
                ["participants", "participantId"],
            ],
            errors="ignore",
            sep="_",
        )
        df2.rename(
            columns={
                "participants_externalContactId": "externalContactId",
                "participants_participantId": "participantId",
            },
            inplace=True,
        )
        for key in ["metrics", "segments", "mediaEndpointStats"]:
            try:
                df2.drop([key], axis=1, inplace=True)
            except KeyError as e:
                logger.info(f"Key {e} not appearing in the response.")

        # LEVEL 3
        # not all levels 3 have the same data, and that creates problems of standardization
        # so I add empty data where it is not available to avoid future errors
        conversations_df = {}
        for i, conversation in enumerate(data_to_merge):
            for j, entry_0 in enumerate(conversation["participants"]):
                for key in list(entry_0.keys()):
                    if key == "sessions":
                        for k, entry_1 in enumerate(entry_0[key]):
                            if "metrics" not in list(entry_1.keys()):
                                conversation["participants"][j][key][k]["metrics"] = []
                            if "segments" not in list(entry_1.keys()):
                                conversation["participants"][j][key][k]["segments"] = []
                            if "mediaEndpointStats" not in list(entry_1.keys()):
                                conversation["participants"][j][key][k][
                                    "mediaEndpointStats"
                                ] = []
            # LEVEL 3 metrics
            df3_1 = pd.json_normalize(
                conversation,
                record_path=["participants", "sessions", "metrics"],
                meta=[
                    ["participants", "sessions", "sessionId"],
                ],
                errors="ignore",
                record_prefix="metrics_",
                sep="_",
            )
            df3_1.rename(
                columns={"participants_sessions_sessionId": "sessionId"}, inplace=True
            )
            # LEVEL 3 segments
            df3_2 = pd.json_normalize(
                conversation,
                record_path=["participants", "sessions", "segments"],
                meta=[
                    ["participants", "sessions", "sessionId"],
                ],
                errors="ignore",
                record_prefix="segments_",
                sep="_",
            )
            df3_2.rename(
                columns={"participants_sessions_sessionId": "sessionId"}, inplace=True
            )
            # LEVEL 3 mediaEndpointStats
            df3_3 = pd.json_normalize(
                conversation,
                record_path=["participants", "sessions", "mediaEndpointStats"],
                meta=[
                    ["participants", "sessions", "sessionId"],
                ],
                errors="ignore",
                record_prefix="mediaEndpointStats_",
                sep="_",
            )
            df3_3.rename(
                columns={"participants_sessions_sessionId": "sessionId"}, inplace=True
            )

            # merging all LEVELs 3 from the same conversation
            dff3_tmp = pd.concat([df3_1, df3_2])
            dff3 = pd.concat([dff3_tmp, df3_3])

            conversations_df.update({i: dff3})

        # NERGING ALL LEVELS
        # LEVELS 3
        for l, key in enumerate(list(conversations_df.keys())):
            if l == 0:
                dff3_f = conversations_df[key]
            else:
                dff3_f = pd.concat([dff3_f, conversations_df[key]])

        # LEVEL 3 with LEVEL 2
        dff2 = pd.merge(dff3_f, df2, how="outer", on=["sessionId"])

        # LEVEL 2 with LEVEL 1
        dff1 = pd.merge(
            df1, dff2, how="outer", on=["externalContactId", "participantId"]
        )

        # LEVEL 1 with LEVEL 0
        dff = pd.merge(df0, dff1, how="outer", on=["conversationId"])

        return dff

    @defaults_from_attrs(
        "report_name",
        "view_type",
        "environment",
        "report_url",
        "post_data_list",
        "start_date",
        "end_date",
        "report_columns",
        "credentials_genesys",
        "conversationId_list",
        "key_list",
    )
    def run(
        self,
        report_name: str = None,
        view_type: str = None,
        view_type_time_sleep: int = 80,
        environment: str = None,
        report_url: str = None,
        post_data_list: List[str] = None,
        end_point: str = "analytics/reporting/exports",
        start_date: str = None,
        end_date: str = None,
        report_columns: List[str] = None,
        conversationId_list: List[str] = None,
        key_list: List[str] = None,
        credentials_genesys: Dict[str, Any] = None,
    ) -> List[str]:
        """
        Task for downloading data from the Genesys API to DF.

        Args:
            report_name (str, optional): The name of this task. Defaults to a general name 'genesys_to_csv'.
            view_type (str, optional): The type of view export job to be created. Defaults to None.
            view_type_time_sleep (int, optional): Waiting time to retrieve data from Genesys API. Defaults to 80.
            post_data_list (List[str], optional): List of string templates to generate json body. Defaults to None.
            end_point (str, optional): Final end point for Genesys connection. Defaults to "analytics/reporting/exports".
            credentials_genesys (Dict[str, Any], optional): Credentials to connect with Genesys API containing CLIENT_ID. Defaults to None.
            start_date (str, optional): Start date of the report. Defaults to None.
            end_date (str, optional): End date of the report. Defaults to None.
            environment (str, optional): Adress of host server. Defaults to None than will be used enviroment from credentials.
            report_url (str, optional): The url of report generated in json response. Defaults to None.
            report_columns (List[str], optional): List of exisiting column in report. Defaults to None.
            conversationId_list (List[str], optional): List of conversationId passed as attribute of GET method. Defaults to None.
            key_list (List[str], optional): List of keys needed to specify the columns in the GET request method. Defaults to None.

        Returns:
            List[str]: List of file names.
        """

        genesys = Genesys(
            report_name=report_name,
            view_type=view_type,
            credentials_genesys=credentials_genesys,
            start_date=start_date,
            end_date=end_date,
            environment=environment,
            report_url=report_url,
            report_columns=report_columns,
        )

        if view_type == "queue_performance_detail_view":
            genesys.genesys_api_connection(
                post_data_list=post_data_list, end_point=end_point
            )

            logger.info(
                f"Waiting {view_type_time_sleep} seconds for caching data in Genesys database."
            )
            # sleep time to allow Genesys generate all exports
            time.sleep(view_type_time_sleep)
            # in order to wait for API POST request add it
            timeout_start = time.time()
            # 30 seconds timeout is minimal but for safety added 60.
            timeout = timeout_start + 60
            # while loop with timeout
            while time.time() < timeout:
                try:
                    genesys.get_reporting_exports_data()
                    urls = [col for col in np.array(genesys.report_data).T][1]
                    if None in urls:
                        logger.warning("Found None object in list of urls.")
                    else:
                        break
                except TypeError:
                    pass

                # There is a need to clear a list before repeating try statement.
                genesys.report_data.clear()

        elif view_type in [
            "agent_performance_summary_view",
            "agent_status_summary_view",
            "agent_status_detail_view",
        ]:
            genesys.genesys_api_connection(
                post_data_list=post_data_list, end_point=end_point
            )
            logger.info(
                f"Waiting for getting data in Genesys database ({view_type_time_sleep} seconds)."
            )
            time.sleep(view_type_time_sleep)

            genesys.get_reporting_exports_data()

        if view_type is not None and end_point == "analytics/reporting/exports":
            failed = [col for col in np.array(genesys.report_data).T][-1]

            if "FAILED" in failed and "COMPLETED" in failed:
                logger.warning("Some reports failed.")

            if len(genesys.report_data) == 0 or np.unique(failed)[0] == "FAILED":
                genesys.delete_all_reporting_exports()
                logger.warning(f"All existing reports were deleted.")
                raise APIError("No exporting reports were generated.")
            elif not None in [col for col in np.array(genesys.report_data).T][1]:
                logger.info("Downloaded the data from the Genesys into the CSV.")
            else:
                logger.info("Succesfully loaded all exports.")

            file_names = genesys.download_all_reporting_exports(
                path=self.local_file_path
            )
            logger.info("Downloaded the data from the Genesys into the CSV.")
            # in order to wait for API GET request call it
            logger.info("Waiting for caching data in Genesys database.")
            time.sleep(20)
            genesys.delete_all_reporting_exports()
            logger.info(f"All existing reports were deleted.")

            return file_names

        elif view_type is None and end_point == "analytics/conversations/details/query":
            if len(post_data_list) > 1:
                logger.error("Not available more than one body for this end-point.")
                raise signals.FAIL(message="Stopping the flow.")

            stop_loop = False
            page_counter = post_data_list[0]["paging"]["pageNumber"]
            merged_data = {}
            while not stop_loop:
                report = genesys.genesys_api_connection(
                    post_data_list=post_data_list, end_point=end_point
                )
                merged_data_frame = self.merge_conversations_dfs(
                    report["conversations"]
                )

                merged_data.update(
                    {post_data_list[0]["paging"]["pageNumber"]: merged_data_frame}
                )

                if page_counter == 1:
                    max_calls = int(np.ceil(report["totalHits"] / 100))
                if page_counter == max_calls:
                    stop_loop = True

                post_data_list[0]["paging"]["pageNumber"] += 1
                page_counter += 1

            for i, key in enumerate(list(merged_data.keys())):
                if i == 0:
                    final_df = merged_data[key]
                else:
                    final_df = pd.concat([final_df, merged_data[key]])

            date = start_date.replace("-", "")
            file_name = f"conversations_detail_{date}".upper() + ".csv"

            final_df.to_csv(
                os.path.join(self.local_file_path, file_name),
                index=False,
                sep="\t",
            )

            logger.info("Downloaded the data from the Genesys into the CSV.")

            return [file_name]

        elif view_type is None and end_point == "conversations":
            data_list = []

            for id in conversationId_list:
                json_file = genesys.genesys_api_connection(
                    post_data_list=post_data_list,
                    end_point=f"{end_point}/{id}",
                    method="GET",
                )
                logger.info(f"Generated webmsg_response for {id}")

                attributes = json_file["participants"][0]["attributes"]
                temp_dict = {
                    key: value for (key, value) in attributes.items() if key in key_list
                }
                temp_dict["conversationId"] = json_file["id"]
                data_list.append(temp_dict)

            df = pd.DataFrame(data_list)
            df = df[df.columns[-1:]].join(df[df.columns[:-1]])

            start = start_date.replace("-", "")
            end = end_date.replace("-", "")

            file_name = f"WEBMESSAGE_{start}-{end}.csv"
            df.to_csv(
                os.path.join(file_name),
                index=False,
                sep="\t",
            )

            logger.info("Downloaded the data from the Genesys into the CSV.")

            return [file_name]



import pandas as pd
import pytest

from viadot.tasks import CustomerGaugeToDF

ENDPOINT = "responses"
CG = CustomerGaugeToDF(endpoint=ENDPOINT)
CUR = 185000
PAGESIZE = 1000


@pytest.mark.looping_api_calls
def test_customer_gauge_to_df_loop():
    df = CG.run(total_load=True, cursor=CUR, pagesize=PAGESIZE)

    assert isinstance(df, pd.DataFrame)
    assert len(df) > PAGESIZE



import sys
from datetime import date, datetime, timedelta
from typing import Any, Dict, List

import pandas as pd
import prefect
import pytz
from O365 import Account
from O365.mailbox import MailBox

from viadot.exceptions import CredentialError
from viadot.sources.base import Source


class Outlook(Source):
    utc = pytz.UTC

    def __init__(
        self,
        mailbox_name: str,
        start_date: str = None,
        end_date: str = None,
        credentials: Dict[str, Any] = None,
        limit: int = 10000,
        request_retries: int = 10,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Outlook connector build for fetching Outlook API source.
        Data are fetched from start to end date range. If start or end date are not provided
        then flow fetched data from yestarday by default.

        Args:
            mailbox_name (str): Mailbox name.
            start_date (str, optional): A filtering start date parameter e.g. "2022-01-01". Defaults to None.
            end_date (str, optional): A filtering end date parameter e.g. "2022-01-02". Defaults to None.
            credentials (Dict[str, Any], optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Application.
            Defaults to None.
            limit (int, optional): Number of fetched top messages. Defaults to 10000.
            request_retries (int, optional): How many times retries to authorizate. Defaults to 10.
        """

        self.logger = prefect.context.get("logger")

        self.credentials = credentials
        if self.credentials is None:
            raise CredentialError("You do not provide credentials!")

        self.request_retries = request_retries
        self.mailbox_name = mailbox_name

        self.start_date = start_date
        self.end_date = end_date
        if self.start_date is not None and self.end_date is not None:
            self.date_range_end_time = datetime.strptime(self.end_date, "%Y-%m-%d")
            self.date_range_start_time = datetime.strptime(self.start_date, "%Y-%m-%d")
        else:
            self.date_range_start_time = date.today() - timedelta(days=1)
            self.date_range_end_time = date.today()

            min_time = datetime.min.time()
            self.date_range_end_time = datetime.combine(
                self.date_range_end_time, min_time
            )
            self.date_range_start_time = datetime.combine(
                self.date_range_start_time, min_time
            )

        self.account = Account(
            (self.credentials["client_id"], self.credentials["client_secret"]),
            auth_flow_type="credentials",
            tenant_id=self.credentials["tenant_id"],
            main_resource=self.mailbox_name,
            request_retries=self.request_retries,
        )
        if self.account.authenticate():
            self.logger.info(f"{self.mailbox_name} Authenticated!")
        else:
            self.logger.info(f"{self.mailbox_name} NOT Authenticated!")

        self.mailbox_obj = self.account.mailbox()

        self.limit = limit

        super().__init__(*args, credentials=self.credentials, **kwargs)

    @staticmethod
    def _get_subfolders(
        folder_structure: dict, folder: MailBox, key_concat: str = ""
    ) -> Dict[str, List]:
        """To retrieve all the subfolder in a MailBox folder.

        Args:
            folder_structure (dict): Dictionary where to save the data.
            folder (MailBox): The MailBox folder from where to extract the subfolders.
            key_concat (str, optional) Previous Mailbox folder structure to add to
                the actual subfolder. Defaults to "".

        Returns:
            Dict[str, List]: `folder_structure` dictionary is returned once it is updated.
        """
        if key_concat:
            tmp_key = key_concat.split("|")
            key_concat = key_concat.replace(f"|{tmp_key[-1]}", "")

        for subfolder in folder.get_folders():
            if subfolder:
                folder_structure.update(
                    {
                        "|".join([key_concat, folder.name, subfolder.name]).lstrip(
                            "|"
                        ): subfolder
                    }
                )

        if folder_structure:
            return folder_structure

    def _get_all_folders(self, mailbox: MailBox) -> dict:
        """To retrieve all folders from a Mailbox object.

        Args:
            mailbox (MailBox): Outlook Mailbox object from where to extract all folder structure.

        Returns:
            dict: Every single folder and subfolder is returned as "parent (sub)folder|(sub)folder": Mailbox.
        """
        dict_folders = self._get_subfolders({}, mailbox)
        final_dict_folders = dict_folders.copy()

        # loop to get all subfolders
        while_dict_folders = {"key": "value"}
        while len(while_dict_folders) != 0:
            while_dict_folders = {}
            for key, value in list(dict_folders.items()):
                tmp_dict_folders = self._get_subfolders({}, value, key_concat=key)
                if tmp_dict_folders:
                    final_dict_folders.update(tmp_dict_folders)
                    while_dict_folders.update(tmp_dict_folders)

            dict_folders = while_dict_folders.copy()

        return final_dict_folders

    def _get_messages_from_mailbox(
        self,
        dict_folder: dict,
        limit: int = 10000,
        address_limit: int = 8000,
        outbox_list: List[str] = ["Sent Items"],
    ) -> list:
        """To retrieve all messages from all the mailboxes passed in the dictionary.

        Args:
            dict_folder (dict): Mailboxes dictionary holder, with the following structure:
                "parent (sub)folder|(sub)folder": Mailbox.
            limit (int, optional): Number of fetched top messages. Defaults to 10000.
            address_limit (int, optional): The maximum number of accepted characters in the sum
                of all email names. Defaults to 8000.
            outbox_list (List[str], optional): List of outbox folders to differenciate between
                Inboxes and Outboxes. Defaults to ["Sent Items"].

        Returns:
            list: A list with all messages from all Mailboxes.
        """
        data = []
        for key, value in list(dict_folder.items()):
            count = 0
            for message in value.get_messages(limit=limit):
                received_time = message.received
                date_obj = datetime.fromisoformat(str(received_time))
                if (
                    self.date_range_start_time.replace(tzinfo=self.utc)
                    < date_obj
                    < self.date_range_end_time.replace(tzinfo=self.utc)
                ):
                    count += 1
                    fetched = message.to_api_data()
                    sender_mail = fetched.get("from", None)
                    if sender_mail is not None:
                        sender_mail = fetched["from"]["emailAddress"]["address"]
                    recivers_list = fetched.get("toRecipients")
                    recivers = " "
                    if recivers_list is not None:
                        for reciver in recivers_list:
                            add_string = f", {reciver['emailAddress']['address']}"
                            if (
                                sum(list(map(len, [recivers, add_string])))
                                >= address_limit
                            ):
                                break
                            else:
                                recivers += add_string

                    categories = " "
                    if message.categories is not None:
                        categories = ", ".join(
                            categories for categories in message.categories
                        )
                    conversation_index = " "
                    if message.conversation_index is not None:
                        conversation_index = message.conversation_index
                    if isinstance(message.subject, str):
                        subject = message.subject.replace("\t", " ")
                    else:
                        subject = message.subject
                    row = {
                        "(sub)folder": value.name,
                        "conversation ID": fetched.get("conversationId"),
                        "conversation index": conversation_index,
                        "categories": categories,
                        "sender": sender_mail,
                        "subject": subject,
                        "recivers": recivers.strip(", "),
                        "received_time": fetched.get("receivedDateTime"),
                        "mail_adress": self.mailbox_name.split("@")[0]
                        .replace(".", "_")
                        .replace("-", "_"),
                    }
                    if any([x.lower() in key.lower() for x in outbox_list]):
                        row["Inbox"] = False
                    else:
                        row["Inbox"] = True

                    data.append(row)
            if count > 0:
                self.logger.info(f"folder: {key.ljust(76, '-')}  messages: {count}")

        return data

    def get_all_mails_to_df(
        self, outbox_list: List[str] = ["Sent Items"]
    ) -> pd.DataFrame:
        """Download all the messages stored in a MailBox folder and subfolders.

        Args:
            outbox_list (List[str], optional): List of outbox folders to differenciate between
                Inboxes and Outboxes. Defaults to ["Sent Items"].

        Returns:
            pd.DataFrame: All messages are stored in a pandas framwork.
        """
        final_dict_folders = self._get_all_folders(self.mailbox_obj)

        data = self._get_messages_from_mailbox(
            final_dict_folders, limit=self.limit, outbox_list=outbox_list
        )

        df = pd.DataFrame(data=data)

        return df

    def to_csv(self, df: pd.DataFrame) -> None:
        file_name = self.mailbox_name.split("@")[0].replace(".", "_").replace("-", "_")
        df.to_csv(f"{file_name}.csv", index=False)



import json
import os
import shutil

import pytest
from prefect.engine import signals

from viadot.tasks import RunGreatExpectationsValidation

CWD = os.getcwd()
GE_PROJECT_PATH = os.path.join(CWD, "expectations_test")
EXPECTATIONS_PATH = os.path.join(GE_PROJECT_PATH, "expectations")


@pytest.fixture(scope="function")
def expectation_suite_pass():
    expectation_suite = {
        "data_asset_type": "Dataset",
        "expectation_suite_name": "failure",
        "expectations": [
            {
                "expectation_type": "expect_table_row_count_to_be_between",
                "kwargs": {"max_value": 3, "min_value": 3},
                "meta": {},
            },
        ],
        "meta": {
            "columns": {
                "All Users": {"description": ""},
                "Date": {"description": ""},
                "M-Site_Better Space: All Landing Page Sessions": {"description": ""},
                "M-site_Accessories: All Landing Page Sessions": {"description": ""},
                "M-site_More Space: All Landing Page Sessions": {"description": ""},
                "M-site_Replacement: All Landing Page Sessions": {"description": ""},
            },
            "great_expectations_version": "0.13.19",
        },
    }

    if not os.path.exists(EXPECTATIONS_PATH):
        os.makedirs(EXPECTATIONS_PATH)

    expectation_suite_path = os.path.join(EXPECTATIONS_PATH, "failure.json")

    with open(expectation_suite_path, "w") as f:
        json.dump(expectation_suite, f)

    yield

    shutil.rmtree(GE_PROJECT_PATH)


@pytest.fixture(scope="function")
def expectation_suite_fail():
    expectation_suite = {
        "data_asset_type": "Dataset",
        "expectation_suite_name": "failure",
        "expectations": [
            {
                "expectation_type": "expect_table_row_count_to_be_between",
                "kwargs": {"max_value": 3, "min_value": 3},
                "meta": {},
            },
            {
                "expectation_type": "expect_column_values_to_be_between",
                "meta": {},
                "kwargs": {"column": "sales", "min_value": 10, "max_value": 50},
            },
        ],
        "meta": {
            "columns": {
                "All Users": {"description": ""},
                "Date": {"description": ""},
                "M-Site_Better Space: All Landing Page Sessions": {"description": ""},
                "M-site_Accessories: All Landing Page Sessions": {"description": ""},
                "M-site_More Space: All Landing Page Sessions": {"description": ""},
                "M-site_Replacement: All Landing Page Sessions": {"description": ""},
            },
            "great_expectations_version": "0.13.19",
        },
    }

    if not os.path.exists(EXPECTATIONS_PATH):
        os.makedirs(EXPECTATIONS_PATH)

    expectation_suite_path = os.path.join(EXPECTATIONS_PATH, "failure.json")

    with open(expectation_suite_path, "w") as f:
        json.dump(expectation_suite, f)

    yield

    shutil.rmtree(GE_PROJECT_PATH)


def test_great_expectations_pass(expectation_suite_pass, DF):
    great_expectations_task = RunGreatExpectationsValidation()

    try:
        great_expectations_task.run(
            df=DF, expectations_path=EXPECTATIONS_PATH, expectation_suite_name="failure"
        )
    except signals.FAIL:
        pytest.fail("Expectation run failed.")


def test_great_expectations_fail(expectation_suite_fail, DF):
    great_expectations_task = RunGreatExpectationsValidation()

    with pytest.raises(signals.FAIL):
        great_expectations_task.run(
            df=DF, expectations_path=EXPECTATIONS_PATH, expectation_suite_name="failure"
        )
